"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

import collections
import glob
import json
import time
import unittest
from logging import getLogger, _loggerClass

from mock import patch, Mock, MagicMock
from mockredis import MockRedis
from redis.exceptions import TimeoutError
from zake.fake_client import FakeClient

from yahoo_panoptes.framework.configuration_manager import *
from yahoo_panoptes.framework.const import RESOURCE_MANAGER_RESOURCE_EXPIRE
from yahoo_panoptes.framework.context import *
from yahoo_panoptes.framework.resources import PanoptesResource, PanoptesResourceSet, PanoptesResourceDSL, \
    PanoptesContext, ParseException, ParseResults, PanoptesResourcesKeyValueStore, PanoptesResourceStore, \
    PanoptesResourceCache, PanoptesResourceError, PanoptesResourceCacheException, PanoptesResourceEncoder
from yahoo_panoptes.framework.utilities.helpers import ordered
from yahoo_panoptes.framework.utilities.key_value_store import PanoptesKeyValueStore

_TIMESTAMP = round(time.time(), 5)


class PanoptesMockRedis(MockRedis):
    def __init__(self, bad_connection=False, timeout=False, **kwargs):
        if bad_connection:
            raise ConnectionError
        super(PanoptesMockRedis, self).__init__(**kwargs)
        self.connection_pool = 'mockredis connection pool'
        self.timeout = timeout

    def get(self, key):
        if self.timeout:
            raise TimeoutError
        else:
            return super(PanoptesMockRedis, self).get(key)

    def set(self, key, value, ex=None, px=None, nx=False, xx=False):
        if self.timeout:
            raise TimeoutError
        else:
            return super(PanoptesMockRedis, self).set(key, value, ex=ex, px=px, nx=nx, xx=xx)


class PanoptesTestKeyValueStore(PanoptesKeyValueStore):
    """
    A custom Key/Value store for the Discovery Manager which uses the namespace demarcated for Discovery Manager
    """

    def __init__(self, panoptes_context):
        super(PanoptesTestKeyValueStore, self).__init__(panoptes_context, 'panoptes_test')


def panoptes_mock_redis_strict_client(**kwargs):
    return PanoptesMockRedis(strict=True)


def panoptes_mock_redis_strict_client_bad_connection(**kwargs):
    return PanoptesMockRedis(bad_connection=True)


def panoptes_mock_redis_strict_client_timeout(**kwargs):
    return PanoptesMockRedis(timeout=True)


def panoptes_mock_kazoo_client(**kwargs):
    return FakeClient()


class MockKafkaClient(object):
    def __init__(self, kafka_brokers):
        self._kafka_brokers = kafka_brokers
        self._brokers = set()

        for broker in self._kafka_brokers:
            self._brokers.add(broker)

    @property
    def brokers(self):
        return self._brokers if len(self._brokers) > 0 else None


class MockZookeeperClient(object):
    def __init__(self):
        self._lock = None

    @property
    def Lock(self):
        return self._lock


class TestResources(unittest.TestCase):
    def setUp(self):
        self.__panoptes_resource_metadata = {'test': 'test', '_resource_ttl': '604800'}
        self.__panoptes_resource = PanoptesResource(resource_site='test', resource_class='test',
                                                    resource_subclass='test',
                                                    resource_type='test', resource_id='test', resource_endpoint='test',
                                                    resource_plugin='test',
                                                    resource_creation_timestamp=_TIMESTAMP,
                                                    resource_ttl=RESOURCE_MANAGER_RESOURCE_EXPIRE)
        self.__panoptes_resource.add_metadata('test', 'test')
        self.__panoptes_resource_set = PanoptesResourceSet()
        mock_valid_timestamp = Mock(return_value=True)
        with patch('yahoo_panoptes.framework.resources.PanoptesValidators.valid_timestamp',
                   mock_valid_timestamp):
            self.__panoptes_resource_set.resource_set_creation_timestamp = _TIMESTAMP
        self.my_dir, self.panoptes_test_conf_file = _get_test_conf_file()

    def test_panoptes_resource(self):
        panoptes_resource_metadata = self.__panoptes_resource_metadata
        panoptes_resource = self.__panoptes_resource

        self.assertIsInstance(panoptes_resource, PanoptesResource)
        self.assertEqual(panoptes_resource.resource_site, 'test')
        self.assertEqual(panoptes_resource.resource_class, 'test')
        self.assertEqual(panoptes_resource.resource_subclass, 'test')
        self.assertEqual(panoptes_resource.resource_type, 'test')
        self.assertEqual(panoptes_resource.resource_id, 'test')
        self.assertEqual(panoptes_resource.resource_endpoint, 'test')
        self.assertEqual(panoptes_resource.resource_metadata, panoptes_resource_metadata)
        self.assertEqual(panoptes_resource.resource_ttl, str(RESOURCE_MANAGER_RESOURCE_EXPIRE))
        self.assertEqual(panoptes_resource, panoptes_resource)
        self.assertFalse(panoptes_resource == '1')
        self.assertIsInstance(str(panoptes_resource), str)

        with self.assertRaises(AssertionError):
            panoptes_resource.add_metadata(None, 'test')
        with self.assertRaises(AssertionError):
            panoptes_resource.add_metadata('', 'test')
        with self.assertRaises(ValueError):
            panoptes_resource.add_metadata('1', 'test')
        with self.assertRaises(ValueError):
            panoptes_resource.add_metadata('key', 'test|')
        with self.assertRaises(AssertionError):
            PanoptesResource(resource_site='', resource_class='test', resource_subclass='test',
                             resource_type='test', resource_id='test', resource_endpoint='test')
        with self.assertRaises(AssertionError):
            PanoptesResource(resource_site='test', resource_class='', resource_subclass='test',
                             resource_type='test', resource_id='test', resource_endpoint='test')
        with self.assertRaises(AssertionError):
            PanoptesResource(resource_site='test', resource_class='test', resource_subclass='',
                             resource_type='test', resource_id='test', resource_endpoint='test')
        with self.assertRaises(AssertionError):
            PanoptesResource(resource_site='test', resource_class='test', resource_subclass='test',
                             resource_type='', resource_id='test', resource_endpoint='test')
        with self.assertRaises(AssertionError):
            PanoptesResource(resource_site='test', resource_class='test', resource_subclass='test',
                             resource_type='test', resource_id='', resource_endpoint='test')
        with self.assertRaises(AssertionError):
            PanoptesResource(resource_site=None, resource_class='test', resource_subclass='test',
                             resource_type='test', resource_id='test', resource_endpoint='test')
        with self.assertRaises(AssertionError):
            PanoptesResource(resource_site='test', resource_class=None, resource_subclass='test',
                             resource_type='test', resource_id='test', resource_endpoint='test')
        with self.assertRaises(AssertionError):
            PanoptesResource(resource_site='test', resource_class='test', resource_subclass=None,
                             resource_type='test', resource_id='test', resource_endpoint='test')
        with self.assertRaises(AssertionError):
            PanoptesResource(resource_site='test', resource_class='test', resource_subclass='test',
                             resource_type=None, resource_id='test', resource_endpoint='test')
        with self.assertRaises(AssertionError):
            PanoptesResource(resource_site='test', resource_class='test', resource_subclass='test',
                             resource_type='test', resource_id=None, resource_endpoint='test')
        with self.assertRaises(AssertionError):
            PanoptesResource(resource_site='test', resource_class='test', resource_subclass='test',
                             resource_type='test', resource_id='test', resource_endpoint=None)

        # Test json and raw representations of PanoptesResource
        panoptes_resource_2 = PanoptesResource(resource_site='test', resource_class='test',
                                               resource_subclass='test',
                                               resource_type='test', resource_id='test', resource_endpoint='test',
                                               resource_creation_timestamp=_TIMESTAMP,
                                               resource_plugin='test')
        self.assertEqual(panoptes_resource_2.resource_creation_timestamp, _TIMESTAMP)
        panoptes_resource_2_json = {
            u'resource_site': u'test',
            u'resource_id': u'test',
            u'resource_class': u'test',
            u'resource_plugin': u'test',
            u'resource_creation_timestamp': _TIMESTAMP,
            u'resource_subclass': u'test',
            u'resource_endpoint': u'test',
            u'resource_metadata': {
                u'_resource_ttl': u'604800'
            },
            u'resource_type': u'test'
        }
        self.assertEqual(ordered(json.loads(panoptes_resource_2.json)), ordered(panoptes_resource_2_json))
        panoptes_resource_2_raw = collections.OrderedDict(
            [('resource_site', 'test'),
             ('resource_class', 'test'),
             ('resource_subclass', 'test'),
             ('resource_type', 'test'),
             ('resource_id', 'test'),
             ('resource_endpoint', 'test'),
             ('resource_metadata', collections.OrderedDict(
                 [('_resource_ttl', '604800')])
              ),
             ('resource_creation_timestamp', _TIMESTAMP),
             ('resource_plugin', 'test')])
        self.assertEqual(panoptes_resource_2.raw, panoptes_resource_2_raw)

        # Test resource creation from dict
        with open('tests/test_resources/input/resource_one.json') as f:
            resource_specs = json.load(f)
        resource_from_json = PanoptesResource.resource_from_dict(resource_specs['resources'][0])
        panoptes_resource_3 = PanoptesResource(resource_site="test_site", resource_class="network",
                                               resource_subclass="test_subclass", resource_type="test_type",
                                               resource_id="test_id_1", resource_endpoint="test_endpoint_1",
                                               resource_plugin="key")
        self.assertEqual(resource_from_json, panoptes_resource_3)

    def test_panoptes_resource_set(self):
        panoptes_resource = self.__panoptes_resource
        panoptes_resource_set = self.__panoptes_resource_set
        self.assertEqual(panoptes_resource_set.add(panoptes_resource), None)
        self.assertEqual(len(panoptes_resource_set), 1)
        self.assertEqual(type(panoptes_resource_set.resources), set)
        self.assertIsInstance(str(panoptes_resource_set), str)
        self.assertIsInstance(iter(panoptes_resource_set), collections.Iterable)
        self.assertEqual(panoptes_resource_set.next(), panoptes_resource)
        self.assertEqual(panoptes_resource_set.remove(panoptes_resource), None)
        self.assertEqual(len(panoptes_resource_set), 0)
        self.assertEqual(panoptes_resource_set.resource_set_creation_timestamp, _TIMESTAMP)
        with self.assertRaises(AssertionError):
            panoptes_resource_set.resource_set_creation_timestamp = 0

        panoptes_resource_2 = PanoptesResource(resource_site='test', resource_class='test',
                                               resource_subclass='test',
                                               resource_type='test', resource_id='test2', resource_endpoint='test',
                                               resource_plugin='test',
                                               resource_ttl=RESOURCE_MANAGER_RESOURCE_EXPIRE,
                                               resource_creation_timestamp=_TIMESTAMP)
        panoptes_resource_set.add(panoptes_resource)
        panoptes_resource_set.add(panoptes_resource_2)
        self.assertEqual(len(panoptes_resource_set.get_resources_by_site()['test']), 2)
        self.assertEqual(panoptes_resource_set.resource_set_schema_version, "0.1")

        panoptes_resource_set_json = {
            u'resource_set_creation_timestamp': _TIMESTAMP,
            u'resource_set_schema_version': u'0.1',
            u'resources': [
                {
                    u'resource_site': u'test',
                    u'resource_class': u'test',
                    u'resource_subclass': u'test',
                    u'resource_type': u'test',
                    u'resource_id': u'test2',
                    u'resource_endpoint': u'test',
                    u'resource_metadata': {
                        u'_resource_ttl': u'604800'
                    },
                    u'resource_creation_timestamp': _TIMESTAMP,
                    u'resource_plugin': u'test'
                },
                {
                    u'resource_site': u'test',
                    u'resource_class': u'test',
                    u'resource_subclass': u'test',
                    u'resource_type': u'test',
                    u'resource_id': u'test',
                    u'resource_endpoint': u'test',
                    u'resource_metadata': {
                        u'_resource_ttl': u'604800',
                        u'test': u'test'
                    },
                    u'resource_creation_timestamp': _TIMESTAMP,
                    u'resource_plugin': u'test'
                }
            ]
        }
        self.assertEqual(ordered(panoptes_resource_set_json), ordered(json.loads(panoptes_resource_set.json)))

    def test_panoptes_resources_key_value_store(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file)
        panoptes_resources_kv_store = PanoptesResourcesKeyValueStore(panoptes_context)
        self.assertEqual(panoptes_resources_kv_store.redis_group, const.RESOURCE_MANAGER_REDIS_GROUP)
        self.assertEqual(panoptes_resources_kv_store.namespace, const.RESOURCE_MANAGER_KEY_VALUE_NAMESPACE)

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    def test_panoptes_resource_store(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                           key_value_store_class_list=[PanoptesTestKeyValueStore])
        with self.assertRaises(Exception):
            PanoptesResourceStore(panoptes_context)

        mock_kv_store = Mock(return_value=PanoptesTestKeyValueStore(panoptes_context))
        with patch('yahoo_panoptes.framework.resources.PanoptesContext.get_kv_store', mock_kv_store):
            panoptes_resource_store = PanoptesResourceStore(panoptes_context)
            panoptes_resource_store.add_resource("test_plugin_signature", self.__panoptes_resource)
            resource_key = "plugin|test|site|test|class|test|subclass|test|type|test|id|test|endpoint|test"
            resource_value = panoptes_resource_store.get_resource(resource_key)
            self.assertEqual(self.__panoptes_resource, resource_value)

            panoptes_resource_2 = PanoptesResource(resource_site='test', resource_class='test',
                                                   resource_subclass='test',
                                                   resource_type='test', resource_id='test2', resource_endpoint='test',
                                                   resource_plugin='test',
                                                   resource_ttl=RESOURCE_MANAGER_RESOURCE_EXPIRE,
                                                   resource_creation_timestamp=_TIMESTAMP)
            panoptes_resource_store.add_resource("test_plugin_signature", panoptes_resource_2)
            self.assertIn(self.__panoptes_resource, panoptes_resource_store.get_resources())
            self.assertIn(panoptes_resource_2, panoptes_resource_store.get_resources())

            panoptes_resource_store.delete_resource("test_plugin_signature", panoptes_resource_2)
            self.assertNotIn(panoptes_resource_2, panoptes_resource_store.get_resources())

            panoptes_resource_3 = PanoptesResource(resource_site='test', resource_class='test',
                                                   resource_subclass='test',
                                                   resource_type='test', resource_id='test3', resource_endpoint='test',
                                                   resource_plugin='test3',
                                                   resource_ttl=RESOURCE_MANAGER_RESOURCE_EXPIRE,
                                                   resource_creation_timestamp=_TIMESTAMP)
            panoptes_resource_store.add_resource("test_plugin_signature", panoptes_resource_3)
            self.assertIn(panoptes_resource_3, panoptes_resource_store.get_resources(site='test', plugin_name='test3'))
            self.assertNotIn(self.__panoptes_resource,
                             panoptes_resource_store.get_resources(site='test', plugin_name='test3'))

            # Test key not found
            mock_find_keys = Mock(
                return_value=['dummy',
                              'plugin|test|site|test|class|test|subclass|test|type|test|id|test|endpoint|test'])
            with patch('yahoo_panoptes.framework.resources.PanoptesKeyValueStore.find_keys',
                       mock_find_keys):
                self.assertEqual(1, len(panoptes_resource_store.get_resources()))

            # Test resource store methods raise correct errors
            mock_get = Mock(side_effect=Exception)
            with patch('yahoo_panoptes.framework.resources.PanoptesKeyValueStore.get', mock_get):
                with self.assertRaises(PanoptesResourceError):
                    panoptes_resource_store.get_resource('test3')

            # Test bad input
            with self.assertRaises(AssertionError):
                panoptes_resource_store.get_resource("")
            with self.assertRaises(AssertionError):
                panoptes_resource_store.get_resource(1)

            with self.assertRaises(AssertionError):
                panoptes_resource_store.add_resource("", panoptes_resource_2)
            with self.assertRaises(AssertionError):
                panoptes_resource_store.add_resource("test_plugin_signature", None)
            with self.assertRaises(AssertionError):
                panoptes_resource_store.add_resource("test_plugin_signature", PanoptesResourceStore(panoptes_context))

            with self.assertRaises(AssertionError):
                panoptes_resource_store.delete_resource("", panoptes_resource_2)
            with self.assertRaises(AssertionError):
                panoptes_resource_store.delete_resource("test_plugin_signature", None)
            with self.assertRaises(AssertionError):
                panoptes_resource_store.delete_resource("test_plugin_signature",
                                                        PanoptesResourceStore(panoptes_context))

            with self.assertRaises(AssertionError):
                panoptes_resource_store.get_resources("", "test_plugin_name")
            with self.assertRaises(AssertionError):
                panoptes_resource_store.get_resources("test_site", "")
            with self.assertRaises(AssertionError):
                panoptes_resource_store.get_resources(1, "test_plugin_name")
            with self.assertRaises(AssertionError):
                panoptes_resource_store.get_resources("test_site", 1)

            # Test non-existent key
            with self.assertRaises(PanoptesResourceError):
                panoptes_resource_store.get_resource('tes')

            mock_set = Mock(side_effect=Exception)
            with patch('yahoo_panoptes.framework.resources.PanoptesKeyValueStore.set', mock_set):
                with self.assertRaises(PanoptesResourceError):
                    panoptes_resource_store.add_resource("test_plugin_signature", panoptes_resource_2)

            mock_delete = Mock(side_effect=Exception)
            with patch('yahoo_panoptes.framework.resources.PanoptesKeyValueStore.delete', mock_delete):
                with self.assertRaises(PanoptesResourceError):
                    panoptes_resource_store.delete_resource("test_plugin_signature", panoptes_resource_2)

            with self.assertRaises(PanoptesResourceError):
                panoptes_resource_store._deserialize_resource("tes", "null")

            with self.assertRaises(PanoptesResourceError):
                panoptes_resource_store._deserialize_resource(resource_key, "null")

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    def test_resource_dsl_parsing(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file)

        test_query = 'resource_class = "network" AND resource_subclass = "load-balancer"'

        test_result = (
            'SELECT resources.*, group_concat(key,"|"), group_concat(value,"|") ' +
            'FROM resources ' +
            'LEFT JOIN resource_metadata ON resources.id = resource_metadata.id ' +
            'WHERE (resources.resource_class = "network" ' +
            'AND resources.resource_subclass = "load-balancer") ' +
            'GROUP BY resource_metadata.id ' +
            'ORDER BY resource_metadata.id'
        )

        panoptes_resource_dsl = PanoptesResourceDSL(test_query, panoptes_context)
        self.assertEqual(panoptes_resource_dsl.sql, test_result)

        # This very long query tests all code paths with the DSL parser:
        test_query = 'resource_class = "network" AND resource_subclass = "load-balancer" OR \
                resource_metadata.os_version LIKE "4%" AND resource_site NOT IN ("test_site") \
                AND resource_endpoint IN ("test1","test2") AND resource_type != "a10" OR ' \
                     'resource_metadata.make NOT LIKE \
                "A10%" AND resource_metadata.model NOT IN ("test1", "test2")'

        test_result = (
            'SELECT resources.*,group_concat(key,"|"),group_concat(value,"|") FROM (SELECT resource_metadata.id ' +
            'FROM resources,resource_metadata WHERE (resources.resource_class = "network" ' +
            'AND resources.resource_subclass = "load-balancer" ' +
            'AND resources.resource_site NOT IN ("test_site") ' +
            'AND resources.resource_endpoint IN ("test1","test2") ' +
            'AND resources.resource_type != "a10" ' +
            'AND ((resource_metadata.key = "os_version" ' +
            'AND resource_metadata.value LIKE "4%")) ' +
            'AND resource_metadata.id = resources.id) ' +
            'UNION SELECT resource_metadata.id ' +
            'FROM resources,resource_metadata WHERE (resource_metadata.key = "make" ' +
            'AND resource_metadata.value NOT LIKE "A10%") ' +
            'AND resource_metadata.id = resources.id ' +
            'INTERSECT SELECT resource_metadata.id ' +
            'FROM resources,resource_metadata WHERE (resource_metadata.key = "model" ' +
            'AND resource_metadata.value NOT IN ("test1","test2")) ' +
            'AND resource_metadata.id = resources.id ' +
            'GROUP BY resource_metadata.id ' +
            'ORDER BY resource_metadata.id) AS filtered_resources, ' +
            'resources, resource_metadata WHERE resources.id = filtered_resources.id ' +
            'AND resource_metadata.id = filtered_resources.id GROUP BY resource_metadata.id')

        panoptes_resource_dsl = PanoptesResourceDSL(test_query, panoptes_context)
        self.assertEqual(panoptes_resource_dsl.sql, test_result)

        panoptes_resource_dsl = PanoptesResourceDSL('resource_site = "local"', panoptes_context)
        self.assertIsInstance(panoptes_resource_dsl.tokens, ParseResults)

        with self.assertRaises(AssertionError):
            PanoptesResourceDSL(None, panoptes_context)
        with self.assertRaises(AssertionError):
            PanoptesResourceDSL('', None)
        with self.assertRaises(AssertionError):
            PanoptesResourceDSL('', panoptes_context)
        with self.assertRaises(ParseException):
            PanoptesResourceDSL('resources_site = local', panoptes_context)


class TestPanoptesResourceEncoder(unittest.TestCase):
    def setUp(self):
        self.__panoptes_resource = PanoptesResource(resource_site='test', resource_class='test',
                                                    resource_subclass='test',
                                                    resource_type='test', resource_id='test', resource_endpoint='test',
                                                    resource_plugin='test',
                                                    resource_creation_timestamp=_TIMESTAMP,
                                                    resource_ttl=RESOURCE_MANAGER_RESOURCE_EXPIRE)

    def test_panoptes_resource_encoder(self):
        resource_encoder = PanoptesResourceEncoder()
        self.assertEqual(resource_encoder.default(set()), list(set()))
        self.assertEqual(resource_encoder.default(self.__panoptes_resource),
                         self.__panoptes_resource.__dict__['_PanoptesResource__data'])
        with self.assertRaises(TypeError):
            resource_encoder.default(dict())


class TestPanoptesResourceCache(unittest.TestCase):
    def setUp(self):
        self.__panoptes_resource = PanoptesResource(resource_site='test', resource_class='test',
                                                    resource_subclass='test',
                                                    resource_type='test', resource_id='test', resource_endpoint='test',
                                                    resource_plugin='test',
                                                    resource_creation_timestamp=_TIMESTAMP,
                                                    resource_ttl=RESOURCE_MANAGER_RESOURCE_EXPIRE)
        self.my_dir, self.panoptes_test_conf_file = _get_test_conf_file()

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    def test_resource_cache(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                           key_value_store_class_list=[PanoptesTestKeyValueStore])

        #  Test PanoptesResourceCache methods when setup_resource_cache not yet called
        panoptes_resource_cache = PanoptesResourceCache(panoptes_context)
        test_query = 'resource_class = "network"'
        with self.assertRaises(PanoptesResourceCacheException):
            panoptes_resource_cache.get_resources(test_query)
        panoptes_resource_cache.close_resource_cache()

        panoptes_resource = self.__panoptes_resource
        panoptes_resource.add_metadata("metadata_key1", "test")
        panoptes_resource.add_metadata("metadata_key2", "test")

        kv = panoptes_context.get_kv_store(PanoptesTestKeyValueStore)
        serialized_key, serialized_value = PanoptesResourceStore._serialize_resource(panoptes_resource)
        kv.set(serialized_key, serialized_value)

        mock_kv_store = Mock(return_value=kv)

        with patch('yahoo_panoptes.framework.resources.PanoptesContext.get_kv_store', mock_kv_store):
            # Test errors when setting up resource cache
            mock_resource_store = Mock(side_effect=Exception)
            with patch('yahoo_panoptes.framework.resources.PanoptesResourceStore', mock_resource_store):
                with self.assertRaises(PanoptesResourceCacheException):
                    panoptes_resource_cache.setup_resource_cache()

            mock_connect = Mock(side_effect=Exception)
            with patch('yahoo_panoptes.framework.resources.sqlite3.connect', mock_connect):
                with self.assertRaises(PanoptesResourceCacheException):
                    panoptes_resource_cache.setup_resource_cache()

            mock_get_resources = Mock(side_effect=Exception)
            with patch('yahoo_panoptes.framework.resources.PanoptesResourceStore.get_resources', mock_get_resources):
                with self.assertRaises(PanoptesResourceCacheException):
                    panoptes_resource_cache.setup_resource_cache()

            # Test basic operations
            panoptes_resource_cache.setup_resource_cache()
            self.assertIsInstance(panoptes_resource_cache.get_resources('resource_class = "network"'),
                                  PanoptesResourceSet)
            self.assertEqual(len(panoptes_resource_cache.get_resources('resource_class = "network"')), 0)
            self.assertIn(panoptes_resource, panoptes_resource_cache.get_resources('resource_class = "test"'))
            self.assertEqual(len(panoptes_resource_cache._cached_resources), 2)

            panoptes_resource_cache.close_resource_cache()

            # Mock PanoptesResourceStore.get_resources to return Resources that otherwise couldn't be constructed:
            mock_resources = PanoptesResourceSet()
            mock_resources.add(panoptes_resource)
            bad_panoptes_resource = PanoptesResource(resource_site='test', resource_class='test',
                                                     resource_subclass='test',
                                                     resource_type='test', resource_id='test2',
                                                     resource_endpoint='test',
                                                     resource_plugin='test',
                                                     resource_creation_timestamp=_TIMESTAMP,
                                                     resource_ttl=RESOURCE_MANAGER_RESOURCE_EXPIRE)
            bad_panoptes_resource.__dict__['_PanoptesResource__data']['resource_metadata']['*'] = "test"
            bad_panoptes_resource.__dict__['_PanoptesResource__data']['resource_metadata']['**'] = "test"
            mock_resources.add(bad_panoptes_resource)
            mock_get_resources = Mock(return_value=mock_resources)
            with patch('yahoo_panoptes.framework.resources.PanoptesResourceStore.get_resources', mock_get_resources):
                panoptes_resource_cache.setup_resource_cache()
                self.assertEqual(len(panoptes_resource_cache.get_resources('resource_class = "test"')), 1)


class TestPanoptesContext(unittest.TestCase):
    def setUp(self):
        self.my_dir, self.panoptes_test_conf_file = _get_test_conf_file()

    def test_context_config_file(self):
        # Test invalid inputs for config_file
        with self.assertRaises(AssertionError):
            PanoptesContext('')

        with self.assertRaises(AssertionError):
            PanoptesContext(1)

        with self.assertRaises(PanoptesContextError):
            PanoptesContext(config_file='non.existent.config.file')

        # Test that the default config file is loaded if no config file is present in the arguments or environment
        with patch('yahoo_panoptes.framework.const.DEFAULT_CONFIG_FILE_PATH', self.panoptes_test_conf_file):
            panoptes_context = PanoptesContext()
            self.assertEqual(panoptes_context.config_object.redis_urls[0].url, 'redis://:password@localhost:6379/0')
            del panoptes_context

        # Test that the config file from environment is loaded, if present
        os.environ[const.CONFIG_FILE_ENVIRONMENT_VARIABLE] = self.panoptes_test_conf_file
        panoptes_context = PanoptesContext()
        self.assertEqual(panoptes_context.config_object.redis_urls[0].url, 'redis://:password@localhost:6379/0')
        del panoptes_context

        #  Test bad config configuration files
        for f in glob.glob(os.path.join(self.my_dir, 'config_files/test_panoptes_config_bad_*.ini')):
            with self.assertRaises(PanoptesContextError):
                print 'Going to load bad configuration file: %s' % f
                PanoptesContext(f)

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    def test_context(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file)
        self.assertIsInstance(panoptes_context, PanoptesContext)
        self.assertEqual(panoptes_context.config_object.redis_urls[0].url, 'redis://:password@localhost:6379/0')
        self.assertEqual(str(panoptes_context.config_object.redis_urls[0]), 'redis://:**@localhost:6379/0')
        self.assertEqual(panoptes_context.config_object.zookeeper_servers, set(['localhost:2181']))
        self.assertEqual(panoptes_context.config_object.kafka_brokers, set(['localhost:9092']))
        self.assertIsInstance(panoptes_context.config_dict, dict)
        self.assertIsInstance(panoptes_context.logger, _loggerClass)
        self.assertIsInstance(panoptes_context.redis_pool, MockRedis)
        with self.assertRaises(AttributeError):
            panoptes_context.kafka_client
        with self.assertRaises(AttributeError):
            panoptes_context.zookeeper_client
        del panoptes_context

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client_bad_connection)
    def test_context_redis_bad_connection(self):
        with self.assertRaises(ConnectionError):
            PanoptesContext(self.panoptes_test_conf_file)

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    def test_context_key_value_store(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                           key_value_store_class_list=[PanoptesTestKeyValueStore])
        self.assertIsInstance(panoptes_context, PanoptesContext)
        kv = panoptes_context.get_kv_store(PanoptesTestKeyValueStore)
        self.assertIsInstance(kv, PanoptesTestKeyValueStore)
        self.assertTrue(kv.set('test', 'test'))
        self.assertEqual(kv.get('test'), 'test')
        self.assertEqual(kv.get('non.existent.key'), None)

        with self.assertRaises(AssertionError):
            kv.set(None, None)
        with self.assertRaises(AssertionError):
            kv.set('test', None)
        with self.assertRaises(AssertionError):
            kv.set(None, 'test')
        with self.assertRaises(AssertionError):
            kv.set(1, 'test')
        with self.assertRaises(AssertionError):
            kv.set('test', 1)

        with self.assertRaises(PanoptesContextError):
            PanoptesContext(self.panoptes_test_conf_file, key_value_store_class_list=[None])
        with self.assertRaises(PanoptesContextError):
            PanoptesContext(self.panoptes_test_conf_file, key_value_store_class_list=['test'])
        with self.assertRaises(PanoptesContextError):
            PanoptesContext(self.panoptes_test_conf_file, key_value_store_class_list=[PanoptesMockRedis])

        panoptes_context = PanoptesContext(self.panoptes_test_conf_file)

        with self.assertRaises(PanoptesContextError):
            panoptes_context.get_kv_store(PanoptesTestKeyValueStore)

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client_timeout)
    def test_context_key_value_store_timeout(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                           key_value_store_class_list=[PanoptesTestKeyValueStore])
        kv = panoptes_context.get_kv_store(PanoptesTestKeyValueStore)
        with self.assertRaises(TimeoutError):
            kv.set('test', 'test')
        with self.assertRaises(TimeoutError):
            kv.get('test')

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    @patch('kazoo.client.KazooClient', panoptes_mock_kazoo_client)
    def test_context_message_bus(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file, create_zookeeper_client=True)
        self.assertIsInstance(panoptes_context, PanoptesContext)
        self.assertIsInstance(panoptes_context.zookeeper_client, FakeClient)

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    @patch('kazoo.client.KazooClient', panoptes_mock_kazoo_client)
    def test_context_del_methods(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                           key_value_store_class_list=[PanoptesTestKeyValueStore],
                                           create_message_producer=False, async_message_producer=False,
                                           create_zookeeper_client=True)
        panoptes_context.__del__()
        with self.assertRaises(AttributeError):
            kv_stores = panoptes_context.__kv_stores
        with self.assertRaises(AttributeError):
            redis_pool = panoptes_context.__redis_pool
        with self.assertRaises(AttributeError):
            message_producer = panoptes_context.__message_producer
        with self.assertRaises(AttributeError):
            kafka_client = panoptes_context.__kafka_client
        with self.assertRaises(AttributeError):
            zookeeper_client = panoptes_context.__zookeeper_client

        panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                           key_value_store_class_list=[PanoptesTestKeyValueStore],
                                           create_message_producer=False, async_message_producer=False,
                                           create_zookeeper_client=True)
        with self.assertRaises(AttributeError):
            del panoptes_context.__kv_stores
            panoptes_context.__del__()
        with self.assertRaises(AttributeError):
            del panoptes_context.__redis_pool
            panoptes_context.__del__()
        with self.assertRaises(AttributeError):
            del panoptes_context.__message_producer
            panoptes_context.__del__()
        with self.assertRaises(AttributeError):
            del panoptes_context.__kafka_client
            panoptes_context.__del__()
        with self.assertRaises(AttributeError):
            del panoptes_context.__zookeeper_client
            panoptes_context.__del__()

    def test_root_logger_error(self):
        mock_get_logger = Mock(side_effect=Exception)
        with patch('yahoo_panoptes.framework.context.logging.getLogger', mock_get_logger):
            with self.assertRaises(PanoptesContextError):
                PanoptesContext(self.panoptes_test_conf_file)

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    @patch('kazoo.client.KazooClient', panoptes_mock_kazoo_client)
    def test_message_producer(self):
        mock_kafka_client = MagicMock(return_value=MockKafkaClient(kafka_brokers={'localhost:9092'}))
        with patch('yahoo_panoptes.framework.context.KafkaClient', mock_kafka_client):
            panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                               create_message_producer=True, async_message_producer=False)

            self.assertIsNotNone(panoptes_context.message_producer)

            #  Test error in message queue producer
            mock_panoptes_message_queue_producer = Mock(side_effect=Exception)
            with patch('yahoo_panoptes.framework.context.PanoptesMessageQueueProducer',
                       mock_panoptes_message_queue_producer):
                with self.assertRaises(PanoptesContextError):
                    PanoptesContext(self.panoptes_test_conf_file,
                                    create_message_producer=True, async_message_producer=True)

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    @patch('kazoo.client.KazooClient', panoptes_mock_kazoo_client)
    def test_get_kafka_client(self):
        mock_kafka_client = MockKafkaClient(kafka_brokers={'localhost:9092'})
        mock_kafka_client_init = Mock(return_value=mock_kafka_client)
        with patch('yahoo_panoptes.framework.context.KafkaClient', mock_kafka_client_init):
            panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                               create_message_producer=True, async_message_producer=False)
            self.assertEqual(panoptes_context.kafka_client, mock_kafka_client)

    def test_get_panoptes_logger(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file)
        assert isinstance(panoptes_context._get_panoptes_logger(), logging.Logger)

        #  Test error raised when instantiating logger fails
        mock_get_child = Mock(side_effect=Exception)
        with patch('yahoo_panoptes.framework.context.PanoptesContext._PanoptesContext__rootLogger.getChild',
                   mock_get_child):
            with self.assertRaises(PanoptesContextError):
                PanoptesContext(self.panoptes_test_conf_file)

    @patch("tests.test_framework.PanoptesTestKeyValueStore.__init__")
    def test_get_kv_store_error(self, mock_init):
        mock_init.side_effect = Exception
        with self.assertRaises(PanoptesContextError):
            PanoptesContext(self.panoptes_test_conf_file,
                            key_value_store_class_list=[PanoptesTestKeyValueStore])

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    @patch('kazoo.client.KazooClient', panoptes_mock_kazoo_client)
    def test_zookeeper_client(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                           create_zookeeper_client=True)
        assert isinstance(panoptes_context.zookeeper_client, FakeClient)

        mock_kazoo_client = Mock(side_effect=Exception)
        with patch('yahoo_panoptes.framework.context.kazoo.client.KazooClient', mock_kazoo_client):
            with self.assertRaises(PanoptesContextError):
                PanoptesContext(self.panoptes_test_conf_file,
                                create_zookeeper_client=True)

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    def test_get_redis_connection(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file)
        #  Test get_redis_connection
        with self.assertRaises(IndexError):
            panoptes_context.get_redis_connection("default", shard=1)
        self.assertIsNotNone(panoptes_context.get_redis_connection("dummy", shard=1))

        #  Test redis shard count error
        self.assertEqual(panoptes_context.get_redis_shard_count('dummy'), 1)
        with self.assertRaises(KeyError):
            panoptes_context.get_redis_shard_count('dummy', fallback_to_default=False)

    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    @patch('kazoo.client.KazooClient', panoptes_mock_kazoo_client)
    def test_get_lock(self):
        panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                           create_zookeeper_client=True)

        #  Test bad input
        with self.assertRaises(AssertionError):
            panoptes_context.get_lock('path/to/node', 1, 1, "identifier")
        with self.assertRaises(AssertionError):
            panoptes_context.get_lock('/path/to/node', 0, 1, "identifier")
        with self.assertRaises(AssertionError):
            panoptes_context.get_lock('/path/to/node', 1, -1, "identifier")
        with self.assertRaises(AssertionError):
            panoptes_context.get_lock('/path/to/node', 1, 1)
        #  Test non-callable listener
        with self.assertRaises(AssertionError):
            panoptes_context.get_lock("/path/to/node", timeout=1, retries=1, identifier="test", listener=object())

        #  Test lock acquisition/release among multiple contenders
        lock = panoptes_context.get_lock("/path/to/node", timeout=1, retries=1, identifier="test")
        self.assertIsNotNone(lock)
        lock.release()

        lock2 = panoptes_context.get_lock("/path/to/node", timeout=1, retries=0, identifier="test")
        self.assertIsNotNone(lock2)

        lock3 = panoptes_context.get_lock("/path/to/node", timeout=1, retries=1, identifier="test")
        self.assertIsNone(lock3)
        lock2.release()

        #  Test adding a listener for the lock once acquired
        lock4 = panoptes_context.get_lock("/path/to/node", timeout=1, retries=1, identifier="test", listener=object)
        self.assertIsNotNone(lock4)

        mock_zookeeper_client = MockZookeeperClient()
        with patch('yahoo_panoptes.framework.context.PanoptesContext.zookeeper_client', mock_zookeeper_client):
            self.assertIsNone(panoptes_context.get_lock("/path/to/node", timeout=5, retries=1, identifier="test"))


class TestPanoptesConfiguration(unittest.TestCase):
    def setUp(self):
        self.my_dir, self.panoptes_test_conf_file = _get_test_conf_file()

    def test_configuration(self):
        logger = getLogger(__name__)

        with self.assertRaises(AssertionError):
            PanoptesConfig(logger=logger)

        mock_config = Mock(side_effect=ConfigObjError)
        with patch('yahoo_panoptes.framework.configuration_manager.ConfigObj', mock_config):
            with self.assertRaises(ConfigObjError):
                PanoptesConfig(logger=logger, conf_file=self.panoptes_test_conf_file)

        test_config = PanoptesConfig(logger=logger, conf_file=self.panoptes_test_conf_file)
        _SNMP_DEFAULTS = {'retries': 1, 'timeout': 5, 'community': 'public', 'proxy_port': 10161,
                          'community_string_key': 'snmp_community_string', 'non_repeaters': 0, 'max_repetitions': 25,
                          'connection_factory_class': 'PanoptesSNMPConnectionFactory', 'port': 10161,
                          'connection_factory_module': 'yahoo_panoptes.framework.utilities.snmp.connection'}
        self.assertEqual(test_config.snmp_defaults, _SNMP_DEFAULTS)
        self.assertSetEqual(test_config.sites, {'local'})

        #  Test exception is raised when plugin_type is not specified in config file
        mock_plugin_types = ['dummy']
        with patch('yahoo_panoptes.framework.configuration_manager.const.PLUGIN_TYPES', mock_plugin_types):
            with self.assertRaises(Exception):
                PanoptesConfig(logger=logger, conf_file=self.panoptes_test_conf_file)


class TestPanoptesRedisConnectionConfiguration(unittest.TestCase):
    def test_basic_operations(self):
        panoptes_redis_connection_config = PanoptesRedisConnectionConfiguration(group="test_group",
                                                                                namespace="test_namespace",
                                                                                shard="test_shard",
                                                                                host="test_host",
                                                                                port=123,
                                                                                db="test_db",
                                                                                password=None)
        assert repr(panoptes_redis_connection_config) == panoptes_redis_connection_config.url


def _get_test_conf_file():
    my_dir = os.path.dirname(os.path.realpath(__file__))
    panoptes_test_conf_file = os.path.join(my_dir, 'config_files/test_panoptes_config.ini')

    return my_dir, panoptes_test_conf_file


if __name__ == '__main__':
    unittest.main()



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module defines metrics and their related abstractions
"""
import json
import re
import threading
from time import time

from six import string_types

from .exceptions import PanoptesBaseException
from .resources import PanoptesResource
from .validators import PanoptesValidators

_VALID_KEY = re.compile(r"^[^\d\W]\w*\Z")


METRICS_TIMESTAMP_PRECISION = 3
METRICS_GROUP_SCHEMA_VERSION = '0.2'


class PanoptesMetricsException(PanoptesBaseException):
    pass


class PanoptesMetricsNullException(PanoptesMetricsException):
    pass


class PanoptesMetricType(object):
    GAUGE, COUNTER = range(2)


METRIC_TYPE_NAMES = dict((getattr(PanoptesMetricType, n), n) for n in dir(PanoptesMetricType) if '_' not in n)


class PanoptesMetricValidators(object):
    @classmethod
    def valid_panoptes_resource(cls, resource):
        return resource and isinstance(resource, PanoptesResource)

    @classmethod
    def valid_panoptes_metric_type(cls, metric_type):
        return metric_type is not None and metric_type in METRIC_TYPE_NAMES

    @classmethod
    def valid_panoptes_metric(cls, metric):
        return metric and isinstance(metric, PanoptesMetric)

    @classmethod
    def valid_panoptes_metric_dimension(cls, metric_dimension):
        return metric_dimension and isinstance(metric_dimension, PanoptesMetricDimension)

    @classmethod
    def valid_panoptes_metrics_group(cls, metrics_group):
        return metrics_group and isinstance(metrics_group, PanoptesMetricsGroup)


class PanoptesMetric(object):
    """
    Representation of a metric monitored by a plugin

    A metric has a name and a corresponding value. It may also have associated dimension names and values

    Args:
        metric_name(str): The name of the metric
        metric_value(float): The value of the metric
        metric_type(int): The type of the metric - valid values are attributes of the PanoptesMetricType class
    """

    def __init__(self, metric_name, metric_value, metric_type, metric_creation_timestamp=None):
        assert PanoptesValidators.valid_nonempty_string(metric_name), 'metric_name must be a non-empty str'
        assert PanoptesValidators.valid_number(metric_value), 'metric_value must be number'
        assert PanoptesMetricValidators.valid_panoptes_metric_type(
                metric_type), 'metric_type must be an attribute of PanoptesMetricType'
        assert (metric_creation_timestamp is None) or PanoptesValidators.valid_number(metric_creation_timestamp), \
            'metric_creation_timestamp should be None or a number'

        if not _VALID_KEY.match(metric_name):
            raise ValueError(
                    'metric name "%s" has to match pattern: (letter|"_") (letter | digit | "_")*' % metric_name)

        self.__data = dict()
        self.__data['metric_creation_timestamp'] = round(metric_creation_timestamp, METRICS_TIMESTAMP_PRECISION) if \
            metric_creation_timestamp is not None else round(time(), METRICS_TIMESTAMP_PRECISION)
        self.__data['metric_name'] = metric_name
        self.__data['metric_value'] = metric_value
        self.__metric_type_raw = metric_type
        self.__data['metric_type'] = METRIC_TYPE_NAMES[metric_type].lower()

    @property
    def metric_name(self):
        """
        The name of the metric

        Returns:
            str: The name of the metric
        """
        return self.__data['metric_name']

    @property
    def metric_value(self):
        """
        The value of the metric

        Returns:
            float: The value of the metric
        """
        return self.__data['metric_value']

    @property
    def metric_timestamp(self):
        """
        The creation timestamp of the metric

        Returns:
            float: The creation timestamp of the metric
        """
        return round(self.__data['metric_creation_timestamp'], METRICS_TIMESTAMP_PRECISION)

    @property
    def metric_type(self):
        return self.__metric_type_raw

    @property
    def json(self):
        """
        The JSON representation of the metric

        Returns:
            str: The JSON representation of the metric
        """
        return json.dumps(self.__data, sort_keys=True)

    def __repr__(self):
        return str(self.__data)

    def __hash__(self):
        return hash(self.__data['metric_name'] + str(self.__data['metric_value']))

    def __str__(self):
        return str(self.metric_name) + '|' + str(self.metric_value) + '|' + str(self.metric_type)

    def __eq__(self, other):
        if not isinstance(other, PanoptesMetric):
            return False

        return self.metric_name == other.metric_name and \
            self.metric_value == other.metric_value and \
            self.metric_type == other.metric_type


class PanoptesMetricDimension(object):
    def __init__(self, name, value):
        assert name and isinstance(name, string_types), (
            'dimension name must be non-empty str or unicode, is type %s' % type(name))
        assert value and isinstance(value, string_types), (
            'dimension value for dimension "%s" must be non-empty str or unicode, is type %s' % (name, type(value)))

        if not _VALID_KEY.match(name):
            raise ValueError(
                    'dimension name "%s" has to match pattern: (letter|"_") (letter | digit | "_")*' % name)

        if '|' in value:
            raise ValueError('dimension value "%s" cannot contain |' % value)

        self.__data = dict()
        self.__data['dimension_name'] = name
        self.__data['dimension_value'] = value

    @property
    def name(self):
        return self.__data['dimension_name']

    @property
    def value(self):
        return self.__data['dimension_value']

    @property
    def json(self):
        return json.dumps(self.__data)

    def __repr__(self):
        return str(self.__data)

    def __hash__(self):
        return hash(self.name + self.value)

    def __str__(self):
        return str(self.name) + '|' + str(self.value)

    def __eq__(self, other):
        if not isinstance(other, PanoptesMetricDimension):
            return False
        return self.name == other.name and self.value == other.value


class PanoptesMetricsGroupEncoder(json.JSONEncoder):
    # https://github.com/PyCQA/pylint/issues/414
    def default(self, o):  # pylint: disable=E0202
        if isinstance(o, set):
            return list(o)
        if isinstance(o, PanoptesResource):
            return o.__dict__['_PanoptesResource__data']
        if isinstance(o, PanoptesMetric):
            return o.__dict__['_PanoptesMetric__data']
        if isinstance(o, PanoptesMetricDimension):
            return o.__dict__['_PanoptesMetricDimension__data']
        return json.JSONEncoder.default(self, o)


class PanoptesMetricsGroup(object):
    def __init__(self, resource, group_type, interval):
        assert PanoptesMetricValidators.valid_panoptes_resource(
                resource), 'resource must be an instance of PanoptesResource'
        assert PanoptesValidators.valid_nonempty_string(
                group_type), 'group_type must be a non-empty string'
        assert PanoptesValidators.valid_nonzero_integer(
                interval), 'interval must a integer greater than zero'

        self.__data = dict()
        self.__metrics_index = {metric_type: list() for metric_type in METRIC_TYPE_NAMES}
        self.__data['metrics_group_type'] = group_type
        self.__data['metrics_group_interval'] = interval
        self.__data['metrics_group_creation_timestamp'] = round(time(), METRICS_TIMESTAMP_PRECISION)
        self.__data['metrics_group_schema_version'] = METRICS_GROUP_SCHEMA_VERSION
        self.__data['resource'] = resource
        self.__data['metrics'] = set()
        self.__data['dimensions'] = set()
        self._data_lock = threading.Lock()

    def copy(self):
        copied_metrics_group = PanoptesMetricsGroup(self.resource, self.group_type, self.interval)
        for metric in self.metrics:
            copied_metrics_group.add_metric(metric)
        for dimension in self.dimensions:
            copied_metrics_group.add_dimension(dimension)

        return copied_metrics_group

    def add_metric(self, metric):
        assert PanoptesMetricValidators.valid_panoptes_metric(metric), 'metric must be an instance of PanoptesMetric'

        if metric.metric_name in self.__metrics_index[metric.metric_type]:
            raise KeyError('Metric name "%s" (type "%s") for metrics group type "%s" already populated' %
                           (metric.metric_name, METRIC_TYPE_NAMES[metric.metric_type], self.group_type))
        self.__data['metrics'].add(metric)
        self.__metrics_index[metric.metric_type].append(metric.metric_name)

    def add_dimension(self, dimension):
        assert PanoptesMetricValidators.valid_panoptes_metric_dimension(dimension), 'dimension must be instance ' \
                                                                                    'of PanoptesMetricDimension'
        with self._data_lock:
            if self.contains_dimension_by_name(dimension.name):
                raise KeyError('Dimension name %s already populated. '
                               'Please use upsert_dimension if you need to update dimensions' % dimension.name)
            else:
                self.__data['dimensions'].add(dimension)

    def get_dimension_by_name(self, dimension_name):
        assert dimension_name and isinstance(dimension_name, string_types), (
            'dimension name must be non-empty str or unicode, is type %s' % type(dimension_name))
        dimension = [x for x in self.__data['dimensions'] if x.name == dimension_name]
        if not dimension:
            return None
        else:
            return dimension[0]

    def contains_dimension_by_name(self, dimension_name):
        assert dimension_name and isinstance(dimension_name, string_types), (
            'dimension name must be non-empty str or unicode, is type %s' % type(dimension_name))
        return dimension_name in [x.name for x in self.__data['dimensions']]

    def delete_dimension_by_name(self, dimension_name):
        assert dimension_name and isinstance(dimension_name, string_types), (
            'dimension name must be non-empty str or unicode, is type %s' % type(dimension_name))
        with self._data_lock:
            if self.contains_dimension_by_name(dimension_name):
                dimension = self.get_dimension_by_name(dimension_name)
                self.__data['dimensions'].remove(dimension)

    def upsert_dimension(self, dimension):
        assert PanoptesMetricValidators.valid_panoptes_metric_dimension(
                dimension), 'dimension must be instance of PanoptesMetricDimension'

        if self.contains_dimension_by_name(dimension.name):
            self.delete_dimension_by_name(dimension.name)
        self.__data['dimensions'].add(dimension)

    @staticmethod
    def flatten_dimensions(dimensions):
        """Changes list of dimensions to dict

        Args:
            dimensions(list): List of dictionaries containing name and value for each dimension

        Returns:
            dict: Key is dimension_name, Value is dimension_value
        """
        return {dimension['dimension_name']: dimension['dimension_value'] for dimension in dimensions}

    @staticmethod
    def flatten_metrics(metrics):
        """Changes list of metrics to nested dict

        Args:
            metrics(list): List of dictionaries containing name, value, type

        Returns:
            dict: Keys are counter, gauge, which then contain a dictionary of the metrics name paired with
            values and timestamps for each name.
        """
        metrics_dict = {'counter': {}, 'gauge': {}}

        for metric in metrics:
            metrics_dict[metric['metric_type']][metric['metric_name']] = \
                {'value': metric['metric_value'], 'timestamp': metric['metric_creation_timestamp']}

        return metrics_dict

    @property
    def resource(self):
        return self.__data['resource']

    @property
    def metrics(self):
        return self.__data['metrics']

    @property
    def dimensions(self):
        return self.__data['dimensions']

    @property
    def group_type(self):
        return self.__data['metrics_group_type']

    @property
    def interval(self):
        return self.__data['metrics_group_interval']

    @property
    def schema_version(self):
        return self.__data['metrics_group_schema_version']

    @property
    def creation_timestamp(self):
        return self.__data['metrics_group_creation_timestamp']

    @property
    def json(self):
        return json.dumps(self.__data, cls=PanoptesMetricsGroupEncoder)

    def __repr__(self):
        return str(self.__data)

    def __hash__(self):
        metrics_string = str()
        dimensions_string = str()

        for metric in frozenset(self.metrics):
            metrics_string += str(metric)

        for dimension in frozenset(self.dimensions):
            dimensions_string += str(dimension)

        return hash(str(self.resource) + metrics_string + dimensions_string)

    def __eq__(self, other):
        if not isinstance(other, PanoptesMetricsGroup):
            return False
        return self.resource == other.resource and self.metrics == other.metrics and self.dimensions == other.dimensions


class PanoptesMetricsGroupSet(object):
    def __init__(self):
        self._metrics_groups = set()

    def add(self, metrics_group):
        assert PanoptesMetricValidators.valid_panoptes_metrics_group(
                metrics_group), 'metrics_group must be an instance of PanoptesMetricsGroup'
        self._metrics_groups.add(metrics_group)

    def remove(self, metrics_group):
        assert PanoptesMetricValidators.valid_panoptes_metrics_group(
                metrics_group), 'metrics_group must be an instance of PanoptesMetricsGroup'
        self._metrics_groups.remove(metrics_group)

    @property
    def metrics_groups(self):
        return self._metrics_groups

    def __add__(self, other):
        if not other or not isinstance(other, PanoptesMetricsGroupSet):
            raise TypeError('Unsupported type for addition: {}'.format(type(other)))

        new_metrics_group_set = PanoptesMetricsGroupSet()
        map(new_metrics_group_set.add, self.metrics_groups)
        map(new_metrics_group_set.add, other.metrics_groups)

        return new_metrics_group_set

    def __iter__(self):
        return iter(self._metrics_groups)

    def next(self):
        return next(iter(self._metrics_groups))

    def __len__(self):
        return len(self._metrics_groups)

    def __repr__(self):
        return repr(self._metrics_groups)


class PanoptesMetricSet(object):
    """
    An (un-ordered) set of PanoptesMetrics
    """

    def __init__(self):
        self.__metrics = set()

    def add(self, metric):
        """
        Add a metric to the set

        Args:
            metric (PanoptesMetric): The metric to add

        Returns:
            None

        """
        assert isinstance(metric, PanoptesMetric), 'metric must be an instance of PanoptesMetric'
        self.__metrics.add(metric)

    def remove(self, metric):
        """
        Remove a metric from the set

        Args:
            metric (PanoptesMetric): The metric to remove

        Returns:
            None

        """
        assert isinstance(metric, PanoptesMetric), 'metric must be an instance of PanoptesMetric'
        self.__metrics.remove(metric)

    @property
    def metrics(self):
        """
        Return the list of metrics in this set

        Returns:
            list: The set of metrics

        """
        return self.__metrics

    def __iter__(self):
        return iter(self.__metrics)

    def next(self):
        """
        Returns the next metric in the set

        Returns:
            PanoptesMetric: The next metric in the set
        """
        return next(iter(self.__metrics))

    def __repr__(self):
        return str(self.__metrics)

    def __len__(self):
        return len(self.__metrics)



import sys
import argparse
import os
import signal
import time
import re
from influxdb import InfluxDBClient
from influxdb.exceptions import InfluxDBClientError

from ...framework import const
from ...framework.context import PanoptesContext
from ...framework.resources import PanoptesResourcesKeyValueStore
from ...framework.utilities.consumer import PanoptesConsumer, make_topic_names_for_all_sites, \
    get_consumer_type_from_name
from ...framework.utilities.helpers import parse_config_file, get_client_id


METRICS_TYPE_SUPPORTED = ['gauge', 'counter']
DEFAULT_CONFIG_FILE = '/home/panoptes/conf/influxdb_consumer.ini'


class PanoptesInfluxDBConsumerContext(PanoptesContext):  # pragma: no cover
    """
    This class implements a PanoptesContext without any KV stores, producers or ZK client
    """
    def __init__(self):
        super(PanoptesInfluxDBConsumerContext, self).__init__(
                key_value_store_class_list=[PanoptesResourcesKeyValueStore],
                create_message_producer=False, async_message_producer=False, create_zookeeper_client=False)


class PanoptesInfluxDBConnection(InfluxDBClient):
    """
    Class to create Influxdb client connection
    """
    def __init__(self, host, port, database, retries, timeout, pool_size):
        super(PanoptesInfluxDBConnection, self).__init__(
            host=host, port=port, database=database, retries=retries, timeout=timeout, pool_size=pool_size)


class PanoptesInfluxDBDefaultTransformer(object):
    """
    This class implements Panoptes Metrics Group to InfluxDB line protocol points transformation
    """
    def __init__(self, metrics_group):
        self._resource = metrics_group['resource']
        self._metrics_group = metrics_group

    @property
    def resource(self):
        """
        Returns:
        dict: The resource for the incoming metrics group
        """
        return self._resource

    @property
    def metrics_group(self):
        """
        Returns:
        dict: The incoming metrics group
        """
        return self._metrics_group

    @property
    def measurement(self):
        """
        This function provides a default implementation of the measurement name for InfluxDB.

        Users should override this if they want to change the application name

        Returns:
            str: The measurement name for InfluxDB
        """
        measurement = self._metrics_group['metrics_group_type']
        influx_measurement_string_regex = re.compile(r'[^0-9a-zA-Z_]+')

        return influx_measurement_string_regex.sub('_', measurement)

    @property
    def timestamp(self):
        """
        This function provides a default implementation of the timestamp - copies in the
        metrics_group_creation_timestamp

        Users should override this if they want to change the timestamp

        Returns:
            float: The Unix epoch timestamp in seconds for the points being emitted
        """
        return int(self._metrics_group['metrics_group_creation_timestamp'])

    @property
    def tag_set(self):
        """
        This function provides a default implementation of the tags needed by InfluxDB. It 'flattens' the incoming
        dimensions into comma separated name/value tag sets. It also adds three 'special' dimensions (resource_endpoint,
        resource_site, resource_class, resource_subclass, resource_type and host) taken from the incoming resource
        information

        Users should override this if they want to change the dimensions

        Returns:
            str: Comma separated name/value tags
        """
        tags = {d['dimension_name']: d['dimension_value'] for d in self.metrics_group['dimensions']}

        tags['resource_class'] = self.resource['resource_class']
        tags['resource_subclass'] = self.resource['resource_subclass']
        tags['resource_type'] = self.resource['resource_type']
        tags['resource_endpoint'] = self.resource['resource_endpoint']
        tags['resource_site'] = self.resource['resource_site']

        tag_set = ','.join("{!s}={!s}".format(key, self._escape_influx_special_char(val))
                           for (key, val) in sorted(tags.items()))

        return tag_set

    @property
    def field_set(self):
        """
        This function provides a default implementation of the fields needed by InfluxDB. It 'flattens' the incoming
        metrics into comma separated name/value field sets.

        Users should override this if they want to change the metrics

        Returns:
            str: comma separated name/value fields
        """
        fields = {d['metric_name'] + '__' + d['metric_type']: d['metric_value'] for d in self.metrics_group['metrics']
                  if d['metric_type'] in METRICS_TYPE_SUPPORTED}

        field_set = ','.join("{!s}={!r}".format(key, val) for (key, val) in sorted(fields.items()))

        return field_set

    @staticmethod
    def _escape_influx_special_char(string):
        """
        Method to escape InfluxDB points line protocol special characters

        Args:
            string: InfluxDB field or tag string

        Returns:
            str: field or tag string with special chars escaped
        """
        influx_special_char = re.compile('([,=\s])')
        return influx_special_char.sub(r'\\\1', string)

    def translate_to_influxdb_points(self):
        """
        Method to construct InfluxDB line protocol formatted string

        Returns:
            str: InfluxDB line protocol formatted string
            (measurement,tag1=text1,tag2=text2 field1=val1,field2=val2 timestamp)
        """

        return '{},{} {} {}'.format(self.measurement, self.tag_set, self.field_set, self.timestamp)


class PanoptesInfluxDBConsumer(object):
    CONFIG_SPEC_FILE = os.path.dirname(os.path.realpath(__file__)) + '/influxdb_consumer_configspec.ini'

    def __init__(self, config_file):
        """
        This is the base class for a Panoptes to InfluxDB consumer. It reads metrics from Kafka and sends them to
        InfluxDB write api.

        This class defines default implementations of converting Panoptes Metric Group to InfluxDB line protocol
        string format.

        measurement,tag1=text1,tag2=text2 field1=val1,field2=val2 timestamp

        It is expected that users would inherit and extend these methods if they wish to intercept
        the data in any way before sending to InfluxDB.

        The emission logic has fault tolerance and retries built in.

        Args:
            config_file (str): The name of config file

        Attributes:
            CONFIG_SPEC_FILE: The configspec file to use with configobj - please don't change this till you know what
            you're doing
        """

        try:
            self._config = parse_config_file(config_file, self.CONFIG_SPEC_FILE)
        except Exception as e:
            sys.exit('Error parsing configuration file: {}'.format(repr(e)))

        try:
            self._panoptes_context = PanoptesInfluxDBConsumerContext()
        except Exception as e:
            sys.exit('Could not create a InfluxDB Context: {}'.format(repr(e)))

        self._logger = self._panoptes_context.logger
        self._install_signal_handlers()
        self.influxdb_connection = None

        self._initialize_influxdb_connection()

        topics = make_topic_names_for_all_sites(self._panoptes_context, self._config['kafka']['queue'])
        client_id = get_client_id(prefix=self._config['kafka']['group_id'])

        self._consumer = PanoptesConsumer(self._panoptes_context,
                                          consumer_type=get_consumer_type_from_name(self._config['kafka']['queue']),
                                          topics=topics,
                                          keys=None,
                                          client_id=client_id,
                                          group=self._config['kafka']['group_id'],
                                          poll_timeout=self._config['kafka']['poll_timeout'],
                                          session_timeout=self._config['kafka']['session_timeout'],
                                          max_poll_records=self._config['kafka']['max_poll_records'],
                                          callback=self._process_message,
                                          validate=False)

        self.influxdb_points = set()
        self._last_emitted = 0
        self.influxdb_points_batch_size = 0

        self._consumer.start_consumer()

    def _initialize_influxdb_connection(self):
        """
        Method to initialize InfluxDB connection and creates database.

        Returns:
            None
        """
        logger = self._logger

        self.influxdb_connection = PanoptesInfluxDBConnection(
            host=self._config['influxdb']['host'],
            port=self._config['influxdb']['port'],
            database=self._config['influxdb']['database'],
            retries=self._config['influxdb']['write_api_connect_retries'],
            timeout=self._config['influxdb']['write_api_connect_timeout'],
            pool_size=self._config['influxdb']['write_api_connection_pool_size']
        )

        try:
            if self.influxdb_connection.ping():
                logger.info('Successfully initialized InfluxDB api connection')
            else:
                logger.error('Failed to initialize InfluxDB api connection. Exiting.')
                sys.exit(1)
        except:
            logger.error('Failed to initialize InfluxDB api connection. Exiting.')
            sys.exit(1)

        database_list = [db_entry['name'] for db_entry in self.influxdb_connection.get_list_database()]

        if self._config['influxdb']['database'] in database_list:
            logger.info('Influxdb database {!r} already created..skipping'.format(self._config['influxdb']['database']))
        else:
            try:
                logger.info('Creating InfluxDB database {!r}'.format(self._config['influxdb']['database']))
                self.influxdb_connection.create_database(self._config['influxdb']['database'])
            except Exception as e:
                logger.error('Failed while creating InfluxDB database {!r}: {}'.
                             format(self._config['influxdb']['database'], repr(e)))
                sys.exit(1)

    @classmethod
    def factory(cls):
        """
        This function parses command line arguments and creates a InfluxDB consumer object with the config file provided
        on the command line

        Returns:
            PanoptesInfluxDBConsumer: InfluxDB consumer object with the config file provided
        """
        parser = argparse.ArgumentParser(description='Consume metrics from Panoptes and send them to InfluxDB')

        parser.add_argument('--config',
                            help='Configuration file to use for the consumer. Default: {}'.format(DEFAULT_CONFIG_FILE),
                            default=DEFAULT_CONFIG_FILE)
        try:
            # Using parse_known_args is a hack to get the tests to work under nose
            # https://stackoverflow.com/questions/28976912/how-to-use-nosetests-in-python-while-also-passing
            # -accepting-arguments-for-argpar
            args = parser.parse_known_args()
        except Exception as e:
            sys.exit('Error parsing command line options or configuration file: {}'.format(repr(e)))

        try:
            return cls(args[0].config)
        except Exception as e:
            sys.exit('Error trying to instantiate class: {}'.format(repr(e)))

    def _clear_metrics(self, current_time):
        self.influxdb_points = set()
        self.influxdb_points_batch_size = 0
        self._last_emitted = current_time

    def _send_one_by_one(self):
        """
        This function attempts to send metrics to InfluxDB one by one, instead of a batch. It skips any
        metrics that fail.
        If it fails to emit *all* metric, it will return False - to retry the points - this is done based on the
        assumption that all emission fails, it is probably a transient InfluxDB write api unavailability

        Returns:
            bool: True if it was able to emit some metrics to InfluxDB, false otherwise
        """
        logger = self._logger
        points_skipped = 0

        logger.warn('Client error trying to send {} points to InfluxDB api, going to send each point individually'.
                    format(len(self.influxdb_points)))

        for point in self.influxdb_points:
            try:
                self.influxdb_connection.write_points([point], time_precision='s', protocol='line')
                logger.info('Successfully sent a point of {} bytes'.format(sys.getsizeof(point)))
            except Exception as e:
                points_skipped += 1
                logger.error('Failed while trying to send point: {}'.format(repr(e)))

        if points_skipped == len(self.influxdb_points):
            logger.error('Unable to emit any metric to InfluxDB api, will retry')
            return False
        else:
            logger.info('Successfully sent {} points to InfluxDB api and failed {} points'.
                        format(len(self.influxdb_points) - points_skipped, points_skipped))
            self._clear_metrics(self.current_time)
            return True

    def _send_to_influxdb(self, point):
        """
        This method attempts to send points to InfluxDB write api in batch and retries incase of failures.
        It calls _send_one_by_one method when it gets response code 400 (unable to parse).

        Args:
            point(str): InfluxDB points string

        Returns:
            bool: True if it was able to emit some metrics to YAMAS2, false otherwise
        """
        logger = self._logger

        self.influxdb_points.add(point)
        self.influxdb_points_batch_size = len(self.influxdb_points)

        time_over_emit_interval = round(self.current_time - self._last_emitted)

        if self.influxdb_points_batch_size >= self._config['influxdb']['write_api_batch_size'] \
                or time_over_emit_interval >= self._config['influxdb']['write_api_max_emit_interval']:

            logger.info('Going to send {} bytes to InfluxDB api ({} points, {}s over emit interval)'.
                        format(sys.getsizeof(self.influxdb_points), len(self.influxdb_points), time_over_emit_interval))

            for retry in range(0, self._config['influxdb']['write_api_commit_retries']):
                try:
                    logger.info('Try..{}'.format(retry))
                    self.influxdb_connection.write_points(list(self.influxdb_points), time_precision='s',
                                                          protocol='line',
                                                          batch_size=self._config['influxdb']['write_api_batch_size'])
                    logger.info('Successfully bulk sent {} points to InfluxDB api'.format(len(self.influxdb_points)))
                    self._clear_metrics(self.current_time)
                    break
                except InfluxDBClientError as e:
                    logger.error('Failed while trying to send {} bytes ({} points): {}'.
                                 format(sys.getsizeof(self.influxdb_points),
                                        len(self.influxdb_points), e.content[0:200]))
                    if e.code == 400:
                        if self._send_one_by_one():
                            break
                        else:
                            continue
                except Exception as error:
                    logger.error('Failed while trying to send {} bytes ({} points): {}'.
                                 format(sys.getsizeof(self.influxdb_points), len(self.influxdb_points), repr(error)))
                    continue

            # Return False to Kafka consumer once we have points above write_api_batch_size in buffer and
            # not able to send *any* of them
            if len(self.influxdb_points) > self._config['influxdb']['write_api_batch_size']:
                logger.warn('Retries failed, will try again after backoff interval {}s'.
                            format(self._config['influxdb']['write_api_fail_backoff_interval']))
                time.sleep(5)
                return False

        return True

    def _process_message(self, key, metrics_group):
        """
        This method is the workhorse of the Infuxdb consumer class. It receives incoming metrics groups from Kafka,
        translates them and emits them to InfluxDB write api

        Args:
            key (str): The key for incoming the message
            metrics_group (dict): The dictionary representation of a metrics group

        Returns:
            bool: True
        """

        logger = self._logger
        self._key = key
        self._metrics_group = metrics_group
        self.current_time = time.time()
        status = False

        try:
            influxdb_point = PanoptesInfluxDBDefaultTransformer(metrics_group).translate_to_influxdb_points()
        except Exception as e:
            logger.error('Failed while transforming influxDB points from Panoptes metrics group {}: {}'.
                         format(metrics_group, repr(e)))
            return True

        if influxdb_point:
            status = self._send_to_influxdb(influxdb_point)

        return status

    def _signal_handler(self, signal_number, _):  # pragma: no cover
        print('Caught {}, shutting down InfluxDB Consumer'.format(const.SIGNALS_TO_NAMES_DICT[signal_number]))
        print('Going to shutdown Kafka consumer')
        try:
            self._consumer.stop_consumer()
        except Exception as e:
            print('Error trying to stop Kafka consumer, shutting down anyway: {}'.format(repr(e)))
        print('Shutdown complete, exiting')
        sys.exit(0)

    def _install_signal_handlers(self):
        signal.signal(signal.SIGTERM, self._signal_handler)
        signal.signal(signal.SIGINT, self._signal_handler)
        signal.signal(signal.SIGHUP, self._signal_handler)


def start():  # pragma: no cover
    PanoptesInfluxDBConsumer.factory()



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements a Resource Manager which reads the discovered resources and add/deletes/updates them in the
Panoptes Resource Store (currently Redis)
"""
import signal
import sys
import time

from ..framework import const
from ..framework.context import PanoptesContext
from ..framework.resources import PanoptesResource, PanoptesResourceSet, PanoptesResourceStore, \
    PanoptesResourcesKeyValueStore
from ..framework.utilities.consumer import PanoptesResourcesConsumer
from ..framework.utilities.helpers import get_client_id

panoptes_context = logger = consumer = resource_store = None


class PanoptesResourceManagerContext(PanoptesContext):
    def __init__(self):
        super(PanoptesResourceManagerContext, self).__init__(
                key_value_store_class_list=[PanoptesResourcesKeyValueStore],
                create_message_producer=False, async_message_producer=False, create_zookeeper_client=True)


def _resource_set_to_dictionary(resource_set):
    """
    Takes a PanoptesResourceSet and returns a dictionary keyed by serialization key of each resource. This is hack since
    Python has no easy way to get objects from sets

    @TODO: This is hack and should be replaced

    Args:
        resource_set (PanoptesResourceSet): The PanoptesResourceSet to convert

    Returns:
        dict: A dictionary which is : {<resource serialization key>: <resource object>]
    """
    return {resource.serialization_key: resource for resource in resource_set}


def handle_resources(plugin_signature, resources):
    resource_site = str(resources['resources'][0]['resource_site'])

    plugin_name = str(plugin_signature.split(':')[0])

    logger.info('For plugin "%s" and site "%s", going to get current set' % (plugin_name, resource_site))

    try:
        current_resource_set = resource_store.get_resources(site=resource_site, plugin_name=plugin_name)
    except:
        logger.exception('Error trying to get current resources for plugin "%s"' % plugin_signature)
        return False

    logger.info('For plugin "%s" and site "%s" current set has %d resources' % (
        plugin_signature, resource_site, len(current_resource_set)))

    new_resource_set = PanoptesResourceSet()

    for resource in resources['resources']:
        new_resource = PanoptesResource.resource_from_dict(resource)
        new_resource_set.add(new_resource)

    current_resource_dict = _resource_set_to_dictionary(current_resource_set)
    new_resource_dict = _resource_set_to_dictionary(new_resource_set)

    logger.info('For plugin "%s" and site "%s", new set has %d resources' % (
        plugin_signature, resource_site, len(new_resource_set)))

    resources_to_delete = current_resource_set.resources.difference(new_resource_set)

    logger.info('For plugin "%s" and site "%s", deleting %d resources' % (
        plugin_signature, resource_site, len(resources_to_delete)))

    for resource in resources_to_delete:
        current_timestamp = current_resource_dict[resource.serialization_key].resource_creation_timestamp
        new_timestamp = resources['resource_set_creation_timestamp']
        if current_timestamp > new_timestamp:
            logger.info('For plugin "%s" and site "%s", resource "%s" has timestamp (%0.2f UTC) greater than of '
                        'new set (%0.2f UTC), skipping deletion' % (
                            plugin_signature, resource_site, str(resource), current_timestamp, new_timestamp))

        else:
            logger.info('For plugin "%s" and site "%s", going to delete resource "%s"' % (
                plugin_signature, resource_site, resource))
            try:
                resource_store.delete_resource(plugin_signature, resource)
            except:
                logger.exception('Error trying to delete resource')
                return False

    resources_to_add = new_resource_set.resources.difference(current_resource_set)

    logger.info('For plugin "%s" and site "%s", adding %d resources' %
                (plugin_signature, resource_site, len(resources_to_add)))

    for resource in resources_to_add:
        logger.debug('Going to add resource "%s"' % resource)
        try:
            resource_store.add_resource(plugin_signature, resource)
        except:
            logger.exception('Error trying to add resource')
            return False

    resources_to_update = current_resource_set.resources.intersection(new_resource_set)

    logger.info('For plugin "%s" and site "%s", updating %d resources' % (
        plugin_signature, resource_site, len(resources_to_update)))

    start_time = time.time()

    resources_skipped = 0
    resources_updated = 0

    for resource in resources_to_update:
        current_timestamp = current_resource_dict[resource.serialization_key].resource_creation_timestamp
        new_timestamp = new_resource_dict[resource.serialization_key].resource_creation_timestamp
        if current_timestamp > new_timestamp:
            logger.info('For plugin "%s" and site "%s", resource "%s" has timestamp (%0.2f UTC) greater than of '
                        'new set (%0.2f UTC), skipping update' % (
                            plugin_signature, resource_site, str(resource), current_timestamp, new_timestamp))
            resources_skipped += 1
        else:
            try:
                new_resource = new_resource_dict[resource.serialization_key]
                logger.debug('Going to update resource "%s"' % new_resource)
                resource_store.add_resource(plugin_signature, new_resource)
                resources_updated += 1
            except:
                logger.exception('Error trying to add resource')
                return False

    end_time = time.time()

    logger.info('For plugin "%s" and site "%s", added/updated %d resources in %0.2f seconds, skipped %d resources' % (
        plugin_signature, resource_site, resources_updated, round(end_time - start_time, 2), resources_skipped))

    return True


def _signal_handler(signal_number, _):
    print('Caught %s, shutting down Resource Manager' % const.SIGNALS_TO_NAMES_DICT[signal_number])

    try:
        if consumer:
            print('Going to shutdown Kafka consumer')
            consumer.stop_consumer()
    except:
        print('Error trying to stop Kafka consumer, shutting down anyway')

    print('Shutdown complete, exiting')
    sys.exit(0)


def _install_signal_handlers():
    signal.signal(signal.SIGTERM, _signal_handler)
    signal.signal(signal.SIGINT, _signal_handler)
    signal.signal(signal.SIGHUP, _signal_handler)


def start():
    global panoptes_context, logger, resource_store, consumer

    try:
        panoptes_context = PanoptesResourceManagerContext()
    except Exception as e:
        sys.exit('Could not create a Panoptes Context: %s' % (str(e)))

    logger = panoptes_context.logger
    _install_signal_handlers()
    client_id = get_client_id(const.RESOURCE_MANAGER_CLIENT_ID_PREFIX)
    resource_store = PanoptesResourceStore(panoptes_context)

    consumer = PanoptesResourcesConsumer(panoptes_context=panoptes_context,
                                         keys=None,
                                         client_id=client_id,
                                         group=const.RESOURCE_MANAGER_KAFKA_GROUP_ID,
                                         poll_timeout=const.RESOURCE_MANAGER_KAFKA_POLL_TIMEOUT,
                                         callback=handle_resources,
                                         max_partition_fetch_bytes=const.RESOURCE_MANAGER_MAX_PARTITION_FETCH_BYTES)

    consumer.start_consumer()


if __name__ == '__main__':
    start()



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements Neighbor EnrichmentGroup defined with schema validator
"""

from ...framework.enrichment import PanoptesEnrichmentSchemaValidator, PanoptesEnrichmentGroup

NEIGHBOR_SCHEMA_NAMESPACE = 'neighbor'


class PanoptesNeighborEnrichmentSchemaValidator(PanoptesEnrichmentSchemaValidator):
    schema = {
        'enrichment_label': {
            'type': 'dict',
            'schema': {
                'description': {'type': 'string', 'required': True},
                'type': {'type': 'string', 'required': True},
                'layer': {'type': 'integer', 'required': True},
                'belongs_to_lag': {'type': 'dict', 'required': True},
                'child_of_vlan': {'type': 'dict', 'required': True},
                'parent_of': {'type': 'dict', 'required': True},
                'ipv6_neighbor': {'type': 'dict', 'required': True},
                'ipv4_neighbor': {'type': 'dict', 'required': True},
                'l2_neighbor': {'type': 'dict', 'required': True},
                'ipv6_address': {'type': 'list', 'required': True},
                'ipv4_address': {'type': 'list', 'required': True},
                'vlans': {'type': 'dict', 'required': True},
                'spanning_tree': {'type': 'dict', 'required': True},
                'mac_address': {'type': 'string', 'required': True}
            }
        }
    }


class PanoptesNeighborEnrichmentGroup(PanoptesEnrichmentGroup):
    def __init__(self, enrichment_ttl, execute_frequency):
        super(PanoptesNeighborEnrichmentGroup, self).__init__(
            namespace=NEIGHBOR_SCHEMA_NAMESPACE,
            schema_validator=PanoptesNeighborEnrichmentSchemaValidator(),
            enrichment_ttl=enrichment_ttl,
            execute_frequency=execute_frequency)



"""
This module implements an abstract Key/Value store based around Redis
"""
import mmh3

from six import string_types

from .. import const


class PanoptesKeyValueStoreException(BaseException):
    pass


class PanoptesKeyValueStoreValidators(object):
    @classmethod
    def valid_kv_store_class(cls, kv_store_class):
        """
        valid_kv_store_class(cls, kv_store_class)

        Checks if the passed class is a subclass of PanoptesKeyValueStore

        Args:
            kv_store_class (class): The class to check

        Returns:
            bool: True if the class is not null and is an subclass of PanoptesPluginInfo
        """
        return kv_store_class and issubclass(kv_store_class, PanoptesKeyValueStore)


class PanoptesKeyValueStore(object):
    redis_group = const.DEFAULT_REDIS_GROUP_NAME
    """
    Interface to the Key/Value store provided by the Panoptes framework

    Args:
        panoptes_context (PanoptesContext): The PanoptesContext to use. The Redis client associated with the context \
        would be used to create the key/value store
        namespace (str): The namespace associated with the key/value store. Namespaces are the mechanism to partition \
        the underlying key/value store
    """

    def __init__(self, panoptes_context, namespace):
        self._namespace = namespace
        self._panoptes_context = panoptes_context
        self._no_of_shards = self._panoptes_context.get_redis_shard_count(self.redis_group)

    def _normalized_key(self, key):
        return const.KV_NAMESPACE_DELIMITER.join([self.namespace, key])

    def _get_redis_shard(self, key):
        """

        Args:
            key (str): The key to hash on to calculate the shard to get data from/to

        Returns:
            redis.StrictRedis: The Redis Connection
        """
        shard_no = mmh3.hash(key, signed=False) % self._no_of_shards
        return self._panoptes_context.get_redis_connection(group=self.redis_group, shard=shard_no)

    @property
    def namespace(self):
        """
        Returns the namespace property of the PanoptesKeyValueStore instance
        """
        return self._namespace

    def get(self, key):
        """
        Get the value associated with the key from the key/value store

        Args:
            key (str): The key whose value should be returned

        Returns:
            str: The value associated with the key. None if the key is not found in the key/value store. Passes \
            through exceptions in case of failure
        """
        assert key and isinstance(key, string_types), 'key must be a non-empty str or unicode'

        return self._get_redis_shard(key).get(self._normalized_key(key))

    def set(self, key, value, expire=604800):
        """
        Set the value associated with the key in the key/value store

        This does an 'upsert' - inserts the key/value if it does not exist and updates the value if the key exists

        Args:
            key (str): The key whose value should be set
            value (str): The value to set
            expire (int): A positive integer that, if sets, would expire the key in the number of seconds specified

        Returns:
            None: Nothing. Passes through exceptions in case of failure

        """
        assert key and isinstance(key, string_types), 'key must be a non-empty str or unicode'
        assert value and isinstance(value, string_types), 'value must be a non-empty str or unicode'
        assert expire is None or (isinstance(expire, int) and expire > 0), \
            'expire must be an integer greater than zero'

        return self._get_redis_shard(key).set(self._normalized_key(key), value, ex=expire)

    def getset(self, key, value, expire=604800):
        """
        Set the value associated with the key in the key/value store and returns the previously set value

        This does an 'upsert' - inserts the key/value if it does not exist and updates the value if the key exists

        Args:
            key (str): The key whose value should be set
            value (str): The value to set
            expire (int): A positive integer that, if sets, would expire the key in the number of seconds specified

        Returns:
            None: Nothing. Passes through exceptions in case of failure

        """
        assert key and isinstance(key, string_types), 'key must be a non-empty str or unicode'
        assert value and isinstance(value, string_types), 'value must be a non-empty str or unicode'
        assert expire is None or (isinstance(expire, int) and expire > 0), \
            'expire must be an integer greater than zero'

        value = self._get_redis_shard(key).getset(self._normalized_key(key), value)
        self._get_redis_shard(key).expire(self._normalized_key(key), expire)
        return value

    def ttl(self, key):
        """
        Return the Time to Live (TTL) in seconds, of the given key.

        Args:
            key (str): The key whose ttl should be obtained

        Returns:
            ttl (int): TTL of the key, in seconds
        """
        assert key and isinstance(key, string_types), 'key must be a non-empty str or unicode'

        return self._get_redis_shard(key).ttl(self._normalized_key(key))

    def find_keys(self, pattern=None):
        """
        Find keys in the key/value store matching the supplied pattern.

        Args:
            pattern (str): The pattern we should match in finding keys

        Returns:
            List<str>: List of keys matching the supplied pattern.
        """
        keys = list()

        # We iterate through all shards to get all keys
        # TODO: Dedup keys
        for shard in range(0, self._no_of_shards):
            redis_shard_connection = self._panoptes_context.get_redis_connection(self.redis_group, shard)
            keys.extend([_.replace(self.namespace + const.KV_NAMESPACE_DELIMITER, '')
                         for _ in redis_shard_connection.scan_iter(match=self._normalized_key(pattern),
                                                                   count=const.KV_STORE_SCAN_ITER_COUNT)])

        return keys

    def delete(self, key):
        assert key and isinstance(key, string_types), 'key must be a non-empty str or unicode'

        return self._get_redis_shard(key).delete(self._normalized_key(key))

    def set_members(self, set_name):
        """
        Get the members associated with the set from the key/value store

        Args:
            set_name (str): The set whose members should be returned

        Returns:
            list: The members associated with the sey. None if the set is not found in the key/value store. Passes \
            through exceptions in case of failure
        """
        assert set_name and isinstance(set_name, string_types), 'set_name must be a non-empty str or unicode'

        return self._get_redis_shard(set_name).smembers(set_name)

    def set_add(self, set_name, member):
        """
        Add a member to the set associated in the the key/value store

        This does an 'upsert' - inserts the set (and member) if it does not exist and updates the member if the set \
        exists

        Args:
            set_name (str): The key whose value should be set
            member (str): The value of the member to add to the set

        Returns:
            None: Nothing. Passes through exceptions in case of failure

        """
        assert set_name and isinstance(set_name, string_types), 'set_name must be a non-empty str or unicode'
        assert member and isinstance(member, string_types), 'member must be a non-empty str or unicode'

        return self._get_redis_shard(set_name).sadd(self._normalized_key(set_name), member)



"""
This module holds various helper functions used throughout the system
"""
import ctypes
import inspect
import logging
import os
import platform
import re
import sys
import threading
import uuid
from _socket import gaierror, herror

import gevent
import ipaddress
from gevent import socket
from gevent.util import wrap_errors

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

from ...framework.exceptions import PanoptesBaseException
from configobj import ConfigObj, ConfigObjError, flatten_errors
from validate import Validator


LOG = logging.getLogger(__name__)


def normalize_plugin_name(plugin_name):
    """
    Return the normalized plugin name so that they can be used safely throughout the system

    The aim of this function is to replace any 'unsafe' characters in the plugin name with characters that would be safe
    across all the supporting services like Redis, ZooKeeper, Kafka etc.

    Args:
        plugin_name (str): The plugin name to normalize

    Returns:
        str: The normalized plugin name

    """
    assert plugin_name and isinstance(plugin_name, str), 'plugin_name must be a non-empty str'
    temp_plugin_name = plugin_name.replace('_', '__')
    normalized_plugin_name = re.sub('[^A-Za-z0-9_]', '_', temp_plugin_name)
    return normalized_plugin_name


def get_module_mtime(module_path):
    """
    Return the unix mtime of a module

    If the module path is a single file, then the mtime of the file is returned
    If the module path is a directory then the mtime of the file with the highest mtime in the directory would be
    returned

    Args:
        module_path (str): The absolute directory path of the module, without the '.py' extension

    Returns:
        int: The mtime of the module

    """
    mtime = 0

    assert isinstance(module_path, str), 'module_path should be str'
    if os.path.isdir(module_path):
        for f in os.listdir(module_path):
            f_time = int(os.path.getmtime(module_path + '/' + f))
            mtime = f_time if f_time > mtime else mtime
    elif os.path.isfile(module_path + '.py'):
        mtime = int(os.path.getmtime(module_path + '.py'))

    return mtime


def resolve_hostnames(hostnames, timeout):
    """
    Do DNS resolution for a given list of hostnames

    This function uses gevent to resolve all the hostnames in *parallel*

    Args:
        hostnames (list): A list of strings
        timeout (int): The number of seconds to wait for resolution of **all** hostnames

    Returns:
        list: A list of (hostname, address) tuples in the same order as the input list of hostnames

    """
    assert isinstance(hostnames, list), 'hostnames should be a list'
    assert isinstance(timeout, int) and timeout > 0, 'timeout should be an int greater than zero'
    jobs = [gevent.spawn(wrap_errors(gaierror, socket.gethostbyname), host) for host in hostnames]
    gevent.joinall(jobs, timeout=timeout)
    addresses = [job.value if not isinstance(job.get(), gaierror) else None for job in jobs]
    results = [(hostnames[i], result) for i, result in enumerate(addresses)]
    return results


def unknown_hostname(ip):
    """
    Returns a custom  hostname for an unresolvable IP

    Args:
        ip (str): The unresolved IP for which to craft the hostname for

    Returns:
        str: The hostname returned is of the format: unknown-x-x-x-x

    """
    return 'unknown-' + re.sub('[.:]', '-', ip)


def get_hostnames(ips, timeout):
    """
    Do DNS resolution for a given list of IPs

    Args:
        ips (list): A list of IPs
        timeout (int): The number of seconds to wait for resolution of **all** IPs

    Returns:
        list: A list of (address, hosname) tuples in the same order as the input list of IPs
    """
    assert isinstance(ips, list), 'ips should be a list'
    assert isinstance(timeout, int) and timeout > 0, 'timeout should be an int greater than zero'
    jobs = [gevent.spawn(wrap_errors((gaierror, herror), socket.gethostbyaddr), ip) for ip in ips]
    gevent.joinall(jobs, timeout=timeout)
    hostnames = [None if isinstance(job.get(), (gaierror, herror)) else job.value for job in jobs]
    results = {
        ips[i]: unknown_hostname(ips[i]) if ((not result) or
                                             (not result[0]) or
                                             result[0].startswith('UNKNOWN'))
        else result[0]
        for i, result in enumerate(hostnames)}
    return results


def get_ip_version(ip):
    # CR: http://stackoverflow.com/questions/11827961/checking-for-ip-addresses
    try:
        socket.inet_aton(ip)
        return 4
    except socket.error:
        pass
    try:
        socket.inet_pton(socket.AF_INET6, ip)
        return 6
    except socket.error:
        pass
    raise ValueError(ip)


def get_hostname():
    """
    Get the hostname of the current host

    Returns:
        str: The hostname

    """
    return platform.node()


def get_os_tid():
    """
    Get the Linux process id associated with the current thread

    Returns:
        int: The process id

    """
    if sys.platform.startswith('linux'):
        return ctypes.CDLL('libc.so.6').syscall(186)
    else:
        # TODO: This is hacky - we need to replace it with something that actually returns the OS thread ID
        return threading._get_ident()


def get_calling_module_name(depth=3):
    """
    Get the name of the 'calling' module

    Args:
        depth (int): How many levels to look 'up' from the current module. Remember that the enclosing 'helpers' \
        module is counted as 1.

    Returns:
        str: The calling module name

    """
    assert isinstance(depth, int), 'depth should be an int'
    frame = inspect.stack()[depth]
    LOG.debug('Got calling frame %r', frame)
    module = inspect.getmodule(frame[0])
    if module:
        return module.__name__


def get_client_id(prefix):
    """
    Generates a client id, based on the given prefix

    Args:
        prefix (str): A non-empty str

    Returns:
        str: The client id which consists of the prefix combined with the hostname and thread id, separated by '_'
    """
    assert prefix and isinstance(prefix, str), 'prefix must be a non-empty str'
    return '_'.join([str(uuid.uuid4()), prefix, get_hostname(), str(get_os_tid())])


class CaptureStdErr(list):
    def __enter__(self):
        self._stderr = sys.stderr
        sys.stderr = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        sys.stderr = self._stderr


class CaptureStdOut(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        sys.stdout = self._stdout


class PanoptesConfigurationParsingError(PanoptesBaseException):
    pass


def parse_config_file(config_file, config_spec_file):
    assert config_file and isinstance(config_file, str), 'config_file must be a non-empty str'
    assert config_spec_file and isinstance(config_spec_file, str), 'config_spec_file must be a non empty str'

    try:
        config = ConfigObj(config_file, configspec=config_spec_file, interpolation='template', file_error=True)
    except IOError as e:
        raise PanoptesConfigurationParsingError('Error reading file: %s' % str(e))
    except ConfigObjError as e:
        raise PanoptesConfigurationParsingError('Error parsing config file "%s": %s' % (config_file, str(e)))

    validator = Validator()
    result = config.validate(validator, preserve_errors=True)

    if result is not True:
        errors = ''
        for (section_list, key, error) in flatten_errors(config, result):
            if key is None:
                errors += 'Section(s) ' + ','.join(section_list) + ' are missing\n'
            else:
                errors += 'The "' + key + '" key in section "' + ','.join(section_list) + '" failed validation\n'

        raise PanoptesConfigurationParsingError('Error parsing the configuration file: %s' % errors)

    return config


def convert_kv_str_to_dict(kv, prefix, kv_delimiter='|', prefix_delimiter='_'):
    """
    This function takes a string encoded list of key/value pairs, separated by a delimiter and turns it into a
    dictionary with key/value pairs with the given prefix and delimiter

    Examples:
        input: (value='metro|WA State|region|Western US|continent|North America|latitude|47.615|longitude|-122.339',
                kv_delimiter='|',
                prefix='geo',
                prefix_delimiter='_')

        output:  {'geo_metro': 'WA State',
                  'geo_region': 'Western US',
                  'geo_continent': 'North America',
                  'geo_latitude': '47.615',
                  'geo_longitude': '-122.339'}


    Args:
        kv (str): The string that encodes the key/value pairs
        prefix (str): The prefix to use for each key in the resultant dictionary
        kv_delimiter (str): The delimiter used to separate the keys and values in the encoded string
        prefix_delimiter (str): The delimiter to use between the prefix and key for the resultant dictionary


    Returns:
        dict: The parsed dictionary of keys and values. Will be an empty dictionary if there isn't a matching number
        of keys and values
    """
    assert isinstance(kv, str) and len(kv) > 0, 'value must be a non-empty string'
    assert isinstance(prefix, str) and len(prefix) > 0, 'prefix must be a non-empty string'
    assert isinstance(kv_delimiter, str) and len(kv_delimiter) > 0, 'kv_delimiter must be a non-empty string'
    assert isinstance(prefix_delimiter, str) and len(
        prefix_delimiter) > 0, 'prefix_delimiter must be a non-empty string'

    fields = kv.split(kv_delimiter)

    kvs = {}
    if (len(fields) % 2) == 0:
        kvs = {"{}{}{}".format(prefix, prefix_delimiter, fields[x * 2]): fields[(x * 2) + 1] for x in
               range(0, len(fields) / 2)}

    return kvs


def convert_celsius_to_fahrenheit(celsius):
    """
    Does exactly what the it says

    Args:
        celsius (int): The input to convert to fahrenheit

    Returns:
        int: The fahrenheit equivalent, rounded to two decimal places

    """

    return round(((celsius * 1.8) + 32), 2)


def ordered(obj):
    """
    Creates and ordered list from a given object. Can handle nested dictionaries and lists

    Args:
        obj (object): The object to order

    Returns:
        object: An ordered list or the original object if it is not a dict or list
    """
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in obj.items())
    if isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    else:
        return obj


def transform_index_ipv6_address(ipv6_str):
    """
    Converts a substring of an SNMP index that contains an IPv6 address into a human readable format.
    Example:
        254.128.0.0.0.0.0.0.14.134.16.255.254.243.135.30 => fe80::e86:10ff:fef3:871e
    Args:
        ipv6_str (str): SNMP index substring containing an IPv6 address.
    Returns:
        str: human readable IPv6 address
    """
    parts = ["{0:02x}".format(int(x)) for x in ipv6_str.split('.')]
    byte_string = ""
    for p, i in enumerate(parts):
        if p % 2 != 0:
            byte_string += '{}{}:'.format(parts[p - 1].lstrip('0'), parts[p])

    return str(ipaddress.ip_address(unicode(byte_string[:-1])))


def transform_octet_to_mac(octet_string):
    """
    Transforms SNMP Octet string to MAC address  separated by ':'
    Args:
        octet_string: SNMP Octet string

    Returns:
        MAC address separated by ':'
    """

    if '\\x' not in repr(octet_string) and octet_string.count('-') == 5:
        mac_address = ':'.join((octet.zfill(2) for octet in octet_string.split('-')))
    else:
        mac_address = ':'.join('{:02x}'.format(ord(field)) for field in octet_string)

    return mac_address.upper()



import json
import os
import unittest
from mock import Mock, patch

from yahoo_panoptes.framework.utilities.helpers import ordered
from tests.plugins.helpers import DiscoveryPluginTestFramework
from yahoo_panoptes.plugins.discovery.plugin_discovery_from_json_file import PluginDiscoveryJSONFile
from yahoo_panoptes.discovery.panoptes_discovery_plugin import PanoptesDiscoveryPluginError

mock_time = Mock()
mock_time.return_value = 1512629517.03121


class TestPluginFromJSONDiscovery(DiscoveryPluginTestFramework, unittest.TestCase):

    plugin = PluginDiscoveryJSONFile()
    path = os.path.dirname(os.path.abspath(__file__))

    def __init__(self, test_name):
        super(TestPluginFromJSONDiscovery, self).__init__(test_name)

    def setUp(self):
        self.plugin = PluginDiscoveryJSONFile()
        super(TestPluginFromJSONDiscovery, self).setUp()

    @patch('yahoo_panoptes.framework.resources.time', mock_time)
    def test_basic_operations(self):
        self._plugin_context.config['main']['config_file'] = \
            os.path.join(os.path.abspath(self.path), 'data/test.json')
        result = self.plugin.run(self._plugin_context)

        self.assertEqual(ordered(json.loads(result.json)), ordered(self._expected_results))

    def test_nonexistent_file(self):
        self._plugin_context.config['main']['config_file'] = \
            os.path.join(os.path.abspath(self.path), 'data/bogus.json')
        with self.assertRaises(PanoptesDiscoveryPluginError):
            self.plugin.run(self._plugin_context)

    def test_bad_resource(self):
        self._plugin_context.config['main']['config_file'] = \
            os.path.join(os.path.abspath(self.path), 'data/bad_resource.json')
        with self.assertRaises(PanoptesDiscoveryPluginError):
            self.plugin.run(self._plugin_context)

    def test_bad_json(self):
        self._plugin_context.config['main']['config_file'] = \
            os.path.join(os.path.abspath(self.path), 'data/bad.json')
        with self.assertRaises(PanoptesDiscoveryPluginError):
            self.plugin.run(self._plugin_context)

    def test_mixed_good_bad_resources(self):
        self._plugin_context.config['main']['config_file'] = \
            os.path.join(os.path.abspath(self.path), 'data/one_bad_resource.json')
        result = self.plugin.run(self._plugin_context)
        self.assertEqual(len(result.resources), 2)

        self._plugin_context.config['main']['config_file'] = \
            os.path.join(os.path.abspath(self.path), 'data/one_bad_resource_in_middle.json')
        result = self.plugin.run(self._plugin_context)
        self.assertEqual(len(result.resources), 3)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module defines the 'base' exception for the Panoptes system
"""


class PanoptesBaseException(Exception):
    """
    The base class for all errors in the Panoptes system
    """
    def __init__(self, *args, **kwargs):
        super(PanoptesBaseException, self).__init__(self, *args, **kwargs)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""


class MockRequestsResponseBadRequest(object):
    status_code = 400
    content = 'Bad Request'


class MockRequestsResponseServerFailure(object):
    status_code = 500
    content = 'Internal Server Failure'


class MockRequestsResponseOK(object):
    status_code = 200
    content = 'OK'



import unittest

from mock import create_autospec

from yahoo_panoptes.framework.const import PLUGIN_AGENT_PLUGIN_EXPIRES_MULTIPLE, \
    PLUGIN_AGENT_PLUGIN_TIME_LIMIT_MULTIPLE
from yahoo_panoptes.framework.plugins.helpers import expires, time_limit
from yahoo_panoptes.framework.plugins.panoptes_base_plugin import PanoptesPluginInfo


class TestFrameworkPlugins(unittest.TestCase):
    def test_helpers(self):
        mock_panoptes_plugin_info = create_autospec(PanoptesPluginInfo)
        mock_panoptes_plugin_info.execute_frequency = 5.75

        #  Assert 'round' is called by testing against non-integer product
        self.assertNotEqual(expires(mock_panoptes_plugin_info),
                            mock_panoptes_plugin_info.execute_frequency * PLUGIN_AGENT_PLUGIN_EXPIRES_MULTIPLE)
        self.assertEqual(expires(mock_panoptes_plugin_info),
                         round(mock_panoptes_plugin_info.execute_frequency * PLUGIN_AGENT_PLUGIN_EXPIRES_MULTIPLE))

        #  Test failure when called on a non-instance of PanoptesPluginInfo
        with self.assertRaises(AssertionError):
            expires(PanoptesPluginInfo)

        #  Assert 'round' is called by testing against non-integer product
        self.assertNotEqual(time_limit(mock_panoptes_plugin_info),
                            mock_panoptes_plugin_info.execute_frequency * PLUGIN_AGENT_PLUGIN_TIME_LIMIT_MULTIPLE)
        self.assertEqual(time_limit(mock_panoptes_plugin_info),
                         round(mock_panoptes_plugin_info.execute_frequency * PLUGIN_AGENT_PLUGIN_TIME_LIMIT_MULTIPLE))

        #  Test failure when called on a non-instance of PanoptesPluginInfo
        with self.assertRaises(AssertionError):
            time_limit(PanoptesPluginInfo)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements classes that can be used by any Plugin Scheduler to setup a Celery App and Celery Beat
"""
import signal
import sys
import threading

from .. import const
from ..validators import PanoptesValidators
from ..context import PanoptesContextValidators
from ..celery_manager import PanoptesCeleryInstance, PanoptesCeleryValidators
from ..utilities.helpers import get_client_id, get_os_tid
from ..utilities.lock import PanoptesLock


class PanoptesPluginScheduler(object):
    """
    This class implements methods to start and manage Celery App and Celery Beat for Plugin Schedulers

    Args:
        panoptes_context(PanoptesContext): The Panoptes Context instance that should be used by the Plugin Scheduler
        plugin_type (str): The type of the plugins the Plugin Scheduler would handle
        plugin_type_display_name (str): The display name that should be used by the Plugin Scheduler in logs and errors
        celery_config (PanoptesCeleryConfig): The Celery Config instance that should be used by the Plugin Scheduler
        lock_timeout (int): The number of seconds to wait before a lock times out and is retried
        plugin_scheduler_task (callable): The callback function that the Plugin Scheduler should call every interval

    Returns:
        None
    """

    def __init__(self, panoptes_context, plugin_type, plugin_type_display_name, celery_config, lock_timeout,
                 plugin_scheduler_task):
        assert PanoptesContextValidators.valid_panoptes_context(
                panoptes_context), 'panoptes_context must be an instance of PanoptesContext'
        assert PanoptesValidators.valid_nonempty_string(plugin_type), 'plugin_type must be a non-empty str'
        assert PanoptesValidators.valid_nonempty_string(plugin_type_display_name), \
            'plugin_type_display_name must be a non-empty str'
        assert PanoptesCeleryValidators.valid_celery_config(
                celery_config), 'celery_config must be an instance of PanoptesCeleryConfig'
        assert PanoptesValidators.valid_nonzero_integer(lock_timeout), 'lock_timeout must be an int greater than zero'
        assert PanoptesValidators.valid_callback(plugin_scheduler_task), 'plugin_scheduler_task must be a callable'

        self._panoptes_context = panoptes_context
        self._config = self._panoptes_context.config_dict
        self._logger = self._panoptes_context.logger
        self._shutdown_plugin_scheduler = threading.Event()
        self._plugin_scheduler_celery_beat_service = None
        self._celery_config = celery_config
        self._celery = None
        self._t = None
        self._lock = None
        self._plugin_type = plugin_type
        self._plugin_type_display_name = plugin_type_display_name
        self._lock_timeout = lock_timeout
        self._plugin_scheduler_task = plugin_scheduler_task

    def start(self):
        """
        This function starts the Plugin Scheduler. It installs signal handlers, acquire an distributed lock and then
        return a Celery application instance

        The flow of the startup process is follows:
        start -> _celery_beat_service_started (starts) -> plugin_scheduler_task_thread

        The plugin_scheduler_task_thread runs the plugin_scheduler_task every "['plugin_type']['plugin_scan_interval']"
        seconds, which comes from the system wide configuration file

        The reason for this slightly convoluted startup is that because the plugin_scheduler_task needs the Celery Beat
        Service instance object so that it can update the schedule periodically and this only available after the
        _celery_beat_service_started callback function is called by Celery Beat

        Returns:
            celery.app: The Celery App instance to be used by the scheduler

        """

        logger = self._logger

        logger.info('%s Plugin Scheduler main thread: OS PID: %d' % (self._plugin_type_display_name, get_os_tid()))

        logger.info('Setting up signal handlers')
        self._install_signal_handlers()

        client_id = get_client_id(const.PLUGIN_CLIENT_ID_PREFIX)
        lock_path = const.PLUGIN_SCHEDULER_LOCK_PATH + '/' + self._plugin_type + '/lock'

        logger.info(
                'Creating lock object for %s Plugin Scheduler under lock path "%s"' % (self._plugin_type, lock_path))
        try:
            self._lock = PanoptesLock(context=self._panoptes_context, path=lock_path, timeout=self._lock_timeout,
                                      retries=0, identifier=client_id)
        except Exception as e:
            sys.exit('Failed to create lock object: %s' % repr(e))

        if self._lock.locked:
            logger.info('Starting Celery Beat Service')
            try:
                self._celery = PanoptesCeleryInstance(self._panoptes_context,
                                                      self._celery_config).celery
                self._celery.conf.update(
                        CELERYBEAT_MAX_LOOP_INTERVAL=self._config[self._plugin_type]['celerybeat_max_loop_interval'])
            except Exception as e:
                logger.error('Error trying to start Celery Beat Service: %s' % str(e))

        return self._celery

    def run(self, sender=None, args=None, **kwargs):
        """
        This function is called after the Celery Beat Service has finished initialization.
        The function (re)installs the signal handlers, since they are overwritten by the Celery Beat Service.
        It stores the reference to the Celery Beat Service instance and starts the Plugin Scheduler thread

        Args:
            sender (celery.beat.Service): The Celery Beat Service instance
            args: Variable length argument list
            **kwargs: Arbitrary keyword argument

        Returns:
            None
        """
        logger = self._logger
        logger.info('Reinstalling signal handlers after Celery Beat Service setup')
        self._install_signal_handlers()
        self._plugin_scheduler_celery_beat_service = sender
        self._t = threading.Thread(target=self._plugin_scheduler_task_thread)
        self._t.start()

    def _plugin_scheduler_task_thread(self):
        """
        This function is the entry point of the Plugin Scheduler thread. It checks if the Plugin Scheduler is shutdown
        mode and if not, then calls the plugin_scheduler_task function every 'plugin_scan_interval'
        seconds

        Returns:
            None

        """
        logger = self._logger
        logger.info('%s Plugin Scheduler Task thread: OS PID: %d' % (self._plugin_type_display_name, get_os_tid()))

        while not self._shutdown_plugin_scheduler.is_set():
            if self._lock.locked:
                try:
                    self._plugin_scheduler_task(self._plugin_scheduler_celery_beat_service)
                except Exception:
                    logger.exception('Error trying to execute plugin scheduler task')
            else:
                logger.warn('%s Plugin Scheduler lock not held, skipping scheduling cycle')
            self._shutdown_plugin_scheduler.wait(self._config[self._plugin_type]['plugin_scan_interval'])

        logger.critical('%s Plugin Scheduler Task thread shutdown' % self._plugin_type_display_name)

    def _install_signal_handlers(self):
        """
        Installs signal handlers for SIGTERM, SIGINT and SIGHUP

        Returns:
            None
        """
        signal.signal(signal.SIGTERM, self._signal_handler)
        signal.signal(signal.SIGINT, self._signal_handler)
        signal.signal(signal.SIGHUP, self._signal_handler)

    def _signal_handler(self, signal_number, _):
        """
        The signal handler addresses two scenarios:
        * The Plugin Scheduler thread is alive: sets an event to shutdown the thread
        * The Plugin Scheduler thread is not alive: this can happen if we have not been able to acquire the lock or
        the if the Plugin Scheduler thread quits unexpectedly. In this case, this handler proceeds to call the function
        to shutdown other services (e.g. Celery Beat Service)

        In case a lock has not yet been acquired, then it also cancels the pending lock acquisition request
        In case it receives a SIGHUP, it calls startup_plugin_scheduler. This would re-create the PanoptesContext -
        essentially, re-reading the configuration

        Args:
            signal_number (int): The received signal number
            _ (frame): Current stack frame object

        Returns:
            None
        """
        # If the Plugin Scheduler is already in the process of shutdown, then do nothing - prevents issues
        # with re-entrancy
        if self._shutdown_plugin_scheduler.is_set():
            print('%s Plugin Scheduler already in the process of shutdown, ignoring %s' %
                  (self._plugin_type_display_name, const.SIGNALS_TO_NAMES_DICT[signal_number]))
            return

        print('Caught %s, shutting down %s Plugin Scheduler' % (
            const.SIGNALS_TO_NAMES_DICT[signal_number], self._plugin_type_display_name))

        shutdown_interval = int(int(self._config[self._plugin_type]['plugin_scan_interval']) * 2)
        print('Shutdown/restart may take up to %s seconds' % shutdown_interval)

        print('Signalling for %s Plugin Scheduler Task Thread to shutdown' % self._plugin_type_display_name)
        self._shutdown_plugin_scheduler.set()

        if (self._t is not None) and (self._t.isAlive()):
            self._t.join()

        if (self._t is None) or (not self._t.isAlive()):
            print('%s Plugin Scheduler Task Thread is not active - shutting down other services' %
                  self._plugin_type_display_name)

        if self._plugin_scheduler_celery_beat_service:
            print('Shutting down Celery Beat Service')
            self._plugin_scheduler_celery_beat_service.stop()

        if self._lock:
            print('Releasing lock')
            self._lock.release()

        print('%s Plugin Scheduler shutdown complete')
        sys.exit()



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements Topology EnrichmentGroup defined with schema validator
"""

from ...framework.enrichment import PanoptesEnrichmentSchemaValidator, PanoptesEnrichmentGroup

SCHEMA_NAMESPACE = 'topology'


class PanoptesTopologyEnrichmentSchemaValidator(PanoptesEnrichmentSchemaValidator):
    schema = {
        'enrichment_label': {
            'type': 'dict',
            'schema': {
                'local_interface': {'type': 'string', 'required': True},
                'local_interface_layer': {'type': 'integer', 'required': True},
                'local_lag_members': {'type': 'list', 'required': True},
                'local_interface_stp_mode': {'type': 'string', 'required': True},
                'local_interface_ip_version': {'type': 'dict', 'required': True},
                'remote_device': {'type': ['string', 'dict'], 'required': True},
                'remote_interface': {'type': 'string', 'required': True},
                'remote_interface_index': {'type': 'string', 'required': True},
                'remote_interface_layer': {'type': 'integer', 'required': True},
                'remote_interface_stp_mode': {'type': 'string', 'required': True},
                'remote_lag_members': {'type': 'list', 'required': True},
                'remote_interface_ip_version': {'type': 'dict', 'required': True},
                'map_type': {'type': 'string', 'required': True}
            }
        }
    }


class PanoptesTopologyEnrichmentGroup(PanoptesEnrichmentGroup):
    def __init__(self, enrichment_ttl, execute_frequency):
        super(PanoptesTopologyEnrichmentGroup, self).__init__(
            namespace=SCHEMA_NAMESPACE,
            schema_validator=PanoptesTopologyEnrichmentSchemaValidator(),
            enrichment_ttl=enrichment_ttl,
            execute_frequency=execute_frequency)



import logging
import unittest
import os
import requests_mock

from mock import patch, Mock


from yahoo_panoptes.consumers.influxdb.consumer import PanoptesInfluxDBConsumer
from yahoo_panoptes.framework.context import PanoptesContext, PanoptesContextError
from tests.test_framework import PanoptesMockRedis
from tests.mock_panoptes_consumer import MockPanoptesConsumer, mock_get_client_id
from influxdb import InfluxDBClient


class MockPanoptesContext(PanoptesContext):
    @patch('redis.StrictRedis', PanoptesMockRedis)
    def __init__(self, config_file='tests/config_files/test_panoptes_config.ini'):
        super(MockPanoptesContext, self).__init__(
                key_value_store_class_list=[],
                create_zookeeper_client=False,
                config_file=config_file,
        )

    @property
    def logger(self):
        return logging.getLogger()


MockPanoptesContextWithException = Mock(side_effect=PanoptesContextError())


class MockInfluxDBClient(InfluxDBClient):
    def create_database(self, dbname):
        return True

    def get_list_database(self):
        return [{u'name': u'db1'}, {u'name': u'db2'}, {u'name': u'db3'}]

    def ping(self):
        return '1.6.2'


class TestPanoptesInfluxDBConsumer(unittest.TestCase):
    @patch('yahoo_panoptes.consumers.influxdb.consumer.PanoptesConsumer', MockPanoptesConsumer)
    @patch('redis.StrictRedis', PanoptesMockRedis)
    @patch('yahoo_panoptes.consumers.influxdb.consumer.PanoptesInfluxDBConsumerContext', MockPanoptesContext)
    @patch('yahoo_panoptes.consumers.influxdb.consumer.get_client_id', mock_get_client_id)
    @patch('yahoo_panoptes.consumers.influxdb.consumer.PanoptesInfluxDBConnection', MockInfluxDBClient)
    @patch('yahoo_panoptes.consumers.influxdb.consumer.DEFAULT_CONFIG_FILE',
           'tests/consumers/influxdb/conf/influxdb_consumer.ini')
    def test_panoptes_influxdb_consumer(self):
        "Default Test"

        output_data_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'output/influx_line00.data')

        output_data = open(output_data_file).read()

        MockPanoptesConsumer.files = [
            'consumers/influxdb/input/metrics_group00.json',
            'consumers/influxdb/input/metrics_group01.json',
            'consumers/influxdb/input/metrics_group02.json'
        ]

        with requests_mock.Mocker() as m:
            m.register_uri(requests_mock.POST, "http://localhost:8086/write", status_code=204)

            PanoptesInfluxDBConsumer.factory()

            self.assertEquals(m.last_request.body, output_data)

    @patch('yahoo_panoptes.consumers.influxdb.consumer.PanoptesConsumer', MockPanoptesConsumer)
    @patch('redis.StrictRedis', PanoptesMockRedis)
    @patch('yahoo_panoptes.consumers.influxdb.consumer.PanoptesInfluxDBConsumerContext', MockPanoptesContextWithException)
    @patch('yahoo_panoptes.consumers.influxdb.consumer.get_client_id', mock_get_client_id)
    @patch('yahoo_panoptes.consumers.influxdb.consumer.PanoptesInfluxDBConnection', MockInfluxDBClient)
    @patch('yahoo_panoptes.consumers.influxdb.consumer.DEFAULT_CONFIG_FILE',
           'tests/consumers/influxdb/conf/influxdb_consumer.ini')
    def test_panoptes_influxdb_consumer02(self):
        """Test with bad PanoptesContext"""
        with self.assertRaises(SystemExit):
            PanoptesInfluxDBConsumer.factory()



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This package contains modules which implement a reusable plugin based framework for the Panoptes system
"""



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements a context object that is passed to each plugin during execution

The context of plugin contains it's configuration, the logger it should use, the key/value store it can use and an
optional, arbitrary data object to be passed to the plugin
"""
import unittest

from mock import patch

from yahoo_panoptes.framework.plugins.context import PanoptesPluginContext, PanoptesPluginWithEnrichmentContext
from yahoo_panoptes.framework.resources import PanoptesContext

from .test_framework import PanoptesTestKeyValueStore, panoptes_mock_redis_strict_client
from .helpers import get_test_conf_file

plugin_conf = {
    'Core': {
        'name': 'Test Plugin',
        'module': 'test_plugin'
    },
    'main': {
        'execute_frequency': '60',
        'resource_filter': 'resource_class = "network"'
    }
}


class TestEnrichment:
    pass


class TestPanoptesPluginContexts(unittest.TestCase):
    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    def setUp(self):
        self.my_dir, self.panoptes_test_conf_file = get_test_conf_file()
        self._panoptes_context = PanoptesContext(self.panoptes_test_conf_file)

    def test_panoptes_plugin_context(self):
        test_kv_store = PanoptesTestKeyValueStore(self._panoptes_context)
        test_secrets_store = PanoptesTestKeyValueStore(self._panoptes_context)
        panoptes_plugin_context = PanoptesPluginContext(self._panoptes_context, "test logger", plugin_conf,
                                                        test_kv_store, test_secrets_store, {'test_key': 'test_value'})

        self.assertEqual(panoptes_plugin_context.logger, self._panoptes_context.logger.getChild("test logger"))
        self.assertEqual(panoptes_plugin_context.config, plugin_conf)
        self.assertEqual(panoptes_plugin_context.kv, test_kv_store)
        self.assertEqual(panoptes_plugin_context.secrets, test_secrets_store)
        self.assertEqual(panoptes_plugin_context.data, {'test_key': 'test_value'})
        self.assertSetEqual(panoptes_plugin_context.sites, self._panoptes_context.config_object.sites)
        self.assertDictEqual(panoptes_plugin_context.snmp, self._panoptes_context.config_object.snmp_defaults)

    def test_panoptes_plugin_with_enrichment_context(self):
        test_kv_store = PanoptesTestKeyValueStore(self._panoptes_context)
        test_secrets_store = PanoptesTestKeyValueStore(self._panoptes_context)
        test_enrichment = TestEnrichment()
        panoptes_plugin_with_enrichment_context = PanoptesPluginWithEnrichmentContext(self._panoptes_context,
                                                                                      "test logger", plugin_conf,
                                                                                      test_kv_store, test_secrets_store,
                                                                                      {'test_key': 'test_value'},
                                                                                      test_enrichment)
        self.assertEqual(panoptes_plugin_with_enrichment_context.enrichment, test_enrichment)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements OperationalEnrichmentGroup defined with schema validator
"""

from ...framework.enrichment import PanoptesEnrichmentSchemaValidator, PanoptesEnrichmentGroup

OPERATIONAL_SCHEMA_NAMESPACE = 'operational'


class PanoptesOperationalEnrichmentSchemaValidator(PanoptesEnrichmentSchemaValidator):
    schema = {
        'enrichment_label': {
            'type': 'dict',
            'schema': {
                'snmpenginetime': {'type': 'integer', 'required': False},  # epoch seconds since last restart
                'device_polling_status': {'type': 'integer', 'required': True},
                'last_updated': {'type': 'integer', 'required': False},  # epoch seconds since last update
            }
        }
    }


class PanoptesOperationalEnrichmentGroup(PanoptesEnrichmentGroup):
    def __init__(self, enrichment_ttl, execute_frequency):
        super(PanoptesOperationalEnrichmentGroup, self).__init__(
            namespace=OPERATIONAL_SCHEMA_NAMESPACE,
            schema_validator=PanoptesOperationalEnrichmentSchemaValidator(),
            enrichment_ttl=enrichment_ttl,
            execute_frequency=execute_frequency)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module contains constants used throughout the Panoptes system
"""

# System Wide Constants
import signal
# <CR> The SIGNALS_TO_NAMES_DICT code comes from Python Standard Library By Example
SIGNALS_TO_NAMES_DICT = dict((getattr(signal, n), n)
                             for n in dir(signal) if n.startswith('SIG') and '_' not in n)
KEY_VALUE_NAMESPACE_PREFIX = 'panoptes:'
DEFAULT_ROOT_LOGGER_NAME = 'panoptes'
DEFAULT_LOG_FORMAT = '[%(asctime)s: %(levelname)s/%(processName)s] %(message)s'
CELERY_LOADER_MODULE = 'celery.utils.imports'
DEFAULT_REDIS_GROUP_NAME = 'default'
KV_STORE_DELIMITER = '|'
KV_STORE_SCAN_ITER_COUNT = 1000
KV_NAMESPACE_DELIMITER = ':'

# Configuration Manager Related Constants
CONFIG_FILE_ENVIRONMENT_VARIABLE = 'PANOPTES_CONFIG_FILE'
DEFAULT_CONFIG_FILE_PATH = '/home/panoptes/conf/panoptes.ini'

# Plugin Scheduler Related Constants
PLUGIN_TYPES = ['discovery', 'polling', 'enrichment']
PLUGIN_EXTENSION = 'panoptes-plugin'
PLUGINS_KEY_VALUE_NAMESPACE = KEY_VALUE_NAMESPACE_PREFIX + 'plugins_kv'
PLUGIN_SCHEDULER_LOCK_PATH = '/panoptes/plugin_scheduler'
PLUGIN_CLIENT_ID_PREFIX = 'plugin'

# Plugin Agent Related Constants
PLUGINS_METADATA_KEY_VALUE_NAMESPACE = KEY_VALUE_NAMESPACE_PREFIX + 'plugins_metadata'
PLUGIN_AGENT_LOCK_PATH = '/panoptes/plugin_agent'
PLUGIN_AGENT_LOCK_ACQUIRE_TIMEOUT = 5
PLUGIN_AGENT_PLUGIN_TIMESTAMPS_EXPIRE = 604800
PLUGIN_AGENT_PLUGIN_EXPIRES_MULTIPLE = 2
PLUGIN_AGENT_PLUGIN_TIME_LIMIT_MULTIPLE = 1.25

# Secrets Manager Related Constants
SECRETS_MANAGER_KEY_VALUE_NAMESPACE = KEY_VALUE_NAMESPACE_PREFIX + 'secrets'

# Discovery Manager Related Constants
DISCOVERY_PLUGIN_SCHEDULER_KEY_VALUE_NAMESPACE = KEY_VALUE_NAMESPACE_PREFIX + 'discovery_plugin_scheduler'
DISCOVERY_PLUGIN_SCHEDULER_CELERY_APP_NAME = 'discovery_plugin_scheduler'
DISCOVERY_PLUGIN_SCHEDULER_CELERY_TASK_PREFIX = 'discovery_plugin_scheduler'
DISCOVERY_PLUGIN_SCHEDULER_LOCK_ACQUIRE_TIMEOUT = 5

# Discovery Plugin Agent Related Constants
DISCOVERY_PLUGIN_AGENT_KEY_VALUE_NAMESPACE = KEY_VALUE_NAMESPACE_PREFIX + 'discovery_plugin_agent_kv'
DISCOVERY_PLUGIN_AGENT_MODULE_NAME = 'yahoo_panoptes.discovery.discovery_plugin_agent.discovery_plugin_task'
DISCOVERY_PLUGIN_AGENT_CELERY_APP_NAME = 'discovery_plugin_agent'
DISCOVERY_PLUGIN_AGENT_PLUGINS_CHILD_LOGGER_NAME = 'discovery_plugins'
DISCOVERY_PLUGIN_AGENT_LOCK_PATH = '/panoptes/discovery/plugin_agent/plugins/lock'
DISCOVERY_PLUGIN_AGENT_LOCK_ACQUIRE_TIMEOUT = 5

# Resource Cache Related Constants
RESOURCE_CACHE_DB_CURSOR_SIZE = 1000
RESOURCE_CACHE_UPDATE_INTERVAL = 300

# Resource Manager Related Constants
RESOURCE_MANAGER_REDIS_GROUP = 'resources'
RESOURCE_MANAGER_KEY_VALUE_NAMESPACE = KEY_VALUE_NAMESPACE_PREFIX + 'resource_manager_kv:resource'
RESOURCE_MANAGER_CLIENT_ID_PREFIX = 'resource_manager'
RESOURCE_MANAGER_KAFKA_GROUP_ID = 'resource_manager_group'
RESOURCE_MANAGER_KAFKA_POLL_TIMEOUT = 15
RESOURCE_MANAGER_RESOURCE_EXPIRE = 604800
RESOURCE_MANAGER_MAX_PARTITION_FETCH_BYTES = 10485760

# Polling Plugin Scheduler Related Constants
POLLING_PLUGIN_SCHEDULER_KEY_VALUE_NAMESPACE = KEY_VALUE_NAMESPACE_PREFIX + 'polling_plugin_scheduler_kv'
POLLING_PLUGIN_SCHEDULER_LOCK_ACQUIRE_TIMEOUT = 5
POLLING_PLUGIN_SCHEDULER_CELERY_APP_NAME = 'polling_plugin_scheduler'
POLLING_PLUGIN_SCHEDULER_CELERY_TASK_PREFIX = 'polling_plugin_task'

# Enrichment Plugin Scheduler Related Constants
ENRICHMENT_PLUGIN_SCHEDULER_KEY_VALUE_NAMESPACE = KEY_VALUE_NAMESPACE_PREFIX + 'enrichment_plugin_scheduler_kv'
ENRICHMENT_PLUGIN_SCHEDULER_LOCK_ACQUIRE_TIMEOUT = 5
ENRICHMENT_PLUGIN_SCHEDULER_CELERY_APP_NAME = 'enrichment_plugin_scheduler'
ENRICHMENT_PLUGIN_SCHEDULER_CELERY_TASK_PREFIX = 'enrichment_plugin_task'

# Polling Plugin Agent Related Constants
POLLING_PLUGIN_AGENT_KEY_VALUE_NAMESPACE = KEY_VALUE_NAMESPACE_PREFIX + 'polling_plugin_agent_kv'
POLLING_PLUGIN_AGENT_MODULE_NAME = 'yahoo_panoptes.polling.polling_plugin_agent.polling_plugin_task'
POLLING_PLUGIN_AGENT_CELERY_APP_NAME = 'polling_plugin_agent'
POLLING_PLUGIN_AGENT_PLUGINS_CHILD_LOGGER_NAME = 'polling_plugins'

# Enrichment Plugin Agent Related Constants
ENRICHMENT_REDIS_GROUP = 'enrichments'
ENRICHMENT_PLUGIN_AGENT_KEY_VALUE_NAMESPACE = KEY_VALUE_NAMESPACE_PREFIX + 'enrichment_plugin_agent_kv'
ENRICHMENT_PLUGIN_AGENT_MODULE_NAME = 'yahoo_panoptes.enrichment.enrichment_plugin_agent.enrichment_plugin_task'
ENRICHMENT_PLUGIN_AGENT_CELERY_APP_NAME = 'enrichment_plugin_agent'
ENRICHMENT_PLUGIN_AGENT_PLUGINS_CHILD_LOGGER_NAME = 'enrichment_plugins'

# Enrichment Plugins Results Related Constants
ENRICHMENT_PLUGIN_RESULTS_KEY_VALUE_NAMESPACE = PLUGINS_KEY_VALUE_NAMESPACE + KV_NAMESPACE_DELIMITER + 'enrichment'

# Metrics processing constants
METRICS_CLIENT_ID_PREFIX = 'metrics_processing'
METRICS_KAFKA_GROUP_ID = 'metrics_processing_group'
METRICS_KAFKA_POLL_TIMEOUT = 5
METRICS_KEY_VALUE_NAMESPACE = KEY_VALUE_NAMESPACE_PREFIX + 'metrics_kv'
METRICS_CONFIDENCE_THRESHOLD = 0.33
METRICS_KV_STORE_TTL = 14400
METRICS_KV_STORE_TTL_MULTIPLE = 3

# Metrics topics related constants
METRICS_RAW_TOPIC_SUFFIX = 'metrics'
METRICS_PROCESSED_TOPIC_SUFFIX = 'processed'
METRICS_TOPIC_NAME_DELIMITER = '-'
METRICS_TOPIC_KEY_DELIMITER = ':'
METRICS_REDIS_GROUP = 'metrics'



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

import unittest
from mock import patch
import logging

from tests.mock_panoptes_consumer import MockPanoptesResourcesConsumer, mock_get_client_id
from tests.test_framework import PanoptesMockRedis
from yahoo_panoptes.framework.context import PanoptesContext
from yahoo_panoptes.framework.resources import PanoptesResourcesKeyValueStore

import yahoo_panoptes.resources.manager


class MockPanoptesContext(PanoptesContext):
    @patch('redis.StrictRedis', PanoptesMockRedis)
    def __init__(self, config_file='tests/config_files/test_panoptes_config.ini'):
        super(MockPanoptesContext, self).__init__(
                key_value_store_class_list=[PanoptesResourcesKeyValueStore],
                create_zookeeper_client=False,
                config_file=config_file
        )

    @property
    def logger(self):
        return logging.getLogger()


class TestResourcesManager(unittest.TestCase):
    @patch('yahoo_panoptes.resources.manager.PanoptesResourceManagerContext', MockPanoptesContext)
    @patch('yahoo_panoptes.resources.manager.PanoptesResourcesConsumer', MockPanoptesResourcesConsumer)
    @patch('yahoo_panoptes.resources.manager.get_client_id', mock_get_client_id)
    def _set_and_get_resources(self, input_files):
        MockPanoptesResourcesConsumer.files = input_files
        yahoo_panoptes.resources.manager.start()
        resource_store = yahoo_panoptes.resources.manager.resource_store
        return resource_store.get_resources()

    def test_resource_manager_initial_addition(self):
        """Tests adding initial resource"""
        resources = self._set_and_get_resources(['test_resources/input/resource_one.json'])
        self.assertEqual(len(resources), 1)

    def test_resource_manager_updation(self):
        """Tests that a resource is updated if the incoming resource has a newer timestamp"""
        resources = self._set_and_get_resources(['test_resources/input/resource_one.json',
                                                 'test_resources/input/resource_one_updated.json'])
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources.resources.pop().resource_creation_timestamp, 1526331464.49)

    def test_resource_manager_stale(self):
        """Tests that a resource is NOT updated if it has an older timestamp"""
        resources = self._set_and_get_resources(['test_resources/input/resource_one_updated.json',
                                                 'test_resources/input/resource_one.json'])

        rl = list(resources)
        print(rl[0])
        print(rl[0].resource_metadata)
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources.resources.pop().resource_creation_timestamp, 1526331464.49)

    def test_resource_manager_deletion(self):
        """Tests that a resource is deleted if it incoming new set does not contain it"""
        resources = self._set_and_get_resources(['test_resources/input/resource_one.json',
                                                 'test_resources/input/resource_two.json'])
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources.resources.pop().resource_id, 'test_id_2')




from yahoo_panoptes_snmp.compat import urepr
from yahoo_panoptes_snmp.utils import strip_non_printable

from ...validators import *


class PanoptesSNMPVariable(object):
    def __init__(self, queried_oid, oid, index, value, snmp_type):
        assert PanoptesValidators.valid_numeric_snmp_oid(str(queried_oid)), \
            'queried_oid must be numeric string with a leading period'
        assert PanoptesValidators.valid_numeric_snmp_oid(str(oid)), 'oid must be numeric string with a leading period'
        assert PanoptesValidators.valid_nonempty_string(str(index)), 'index must be None or a non-empty string'
        assert PanoptesValidators.valid_nonempty_string(str(snmp_type)), 'snmp_type must be a non-empty string'

        self._queried_oid = queried_oid
        self._index = self._normalize_index(oid, index, queried_oid)
        self._value = value
        self._snmp_type = snmp_type

    def __repr__(self):
        printable_value = strip_non_printable(self._value)
        return (
            "<{0} value={1} (oid={2}, index={3}, snmp_type={4})>".format(
                    self.__class__.__name__,
                    urepr(printable_value), urepr(self.oid),
                    urepr(self._index), urepr(self._snmp_type)
            )
        )

    @property
    def oid(self):
        return self._queried_oid

    @property
    def index(self):
        return self._index

    @property
    def value(self):
        return self._value

    @property
    def snmp_type(self):
        return self._snmp_type

    @staticmethod
    def _normalize_index(oid, index, queried_oid):
        """
        Parses the provided oid to return both the prefix and the index

        Args:
            oid (str): The OID (as a dotted numerical string) to parse
            index (str): The index, if known
            queried_oid (str): The OID (as a dotted numerical string) which was queried
        Returns:
            tuple: the parsed prefix and index

        """
        prefix = oid[:len(queried_oid)]
        # TODO: This is hacky and should probably be fixed in the normalize_oid function of easysnmp
        if oid != prefix:
            index = oid[len(queried_oid) + 1:] + '.' + index

        return index



from yahoo_panoptes.plugins.enrichment.interface.plugin_enrichment_interface import PluginEnrichmentInterface

_PORT_SPEED_TABLE = {'Gi': 10 ** 9, 'Te': 10 ** 10}


class PluginEnrichmentCiscoInterface(PluginEnrichmentInterface):
    """
    InterfaceEnrichment class for Cisco devices.
    """
    def get_port_speed(self, index):
        """
        Gets the port speed enrichment value for the specified interface associated with the provided index

        Args:
            index (int): The index to look up the associated interface in self._interface_table

        Returns:
            integer: The port speed, based upon the interface name or copied from the configured_speed enrichment in
                     the default case
        """
        if self._interface_table[index]['interface_name'] is None:
            self._interface_table[index]['interface_name'] = self.get_interface_name(index)
        for port_speed_indicator in _PORT_SPEED_TABLE.keys():
            if self._interface_table[index]['interface_name'].startswith(port_speed_indicator):
                return _PORT_SPEED_TABLE[port_speed_indicator]
        return super(PluginEnrichmentCiscoInterface, self).get_port_speed(index)

    def get_parent_interface_name(self, index):
        """
        Gets the parent interface name for the interface associated with the provided index

        Args:
            index (int): The index used to look up the associated interface in self._interface_table

        Returns:
            string: The name of the parent interface, or self._MISSING_VALUE_STRING if the interface has no parent.
                    For Cisco devices, this is everything to the left of the '.' in the interface name, if a '.' is
                    present.
        """
        interface_name = self.get_interface_name(index)
        if '.' in interface_name:
            parent_interface_name = interface_name.split('.')[0]
            return parent_interface_name
        else:
            return self._MISSING_VALUE_STRING
